#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import json
import fnmatch
import logging
import re
from typing import Optional, Set

from .base import GroupCommand, dump_report_data, report_output_parser
from .enterprise import EnterpriseCommand
from .. import api, constants
from ..error import CommandError
from ..params import KeeperParams
from ..proto import enterprise_pb2

distributor_info_parser = argparse.ArgumentParser(prog='distributor info', parents=[report_output_parser])
distributor_info_parser.add_argument('--reload', dest='reload', action='store_true', help='reload distributors')
distributor_info_parser.add_argument('--mc-details', dest='mc_details', action='store_true', help='Display MC details')
distributor_info_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='Display verbose information')

distributor_msp_info_parser = argparse.ArgumentParser(prog='distributor msp-info', parents=[report_output_parser])
distributor_msp_info_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='Display verbose information')
distributor_msp_info_parser.add_argument('msp', action='store', metavar='MSP_NAME',
                                         help='Managed Company Provider identifier (name or ID).')


distributor_license_parser = argparse.ArgumentParser(prog='distributor license')
distributor_license_parser.add_argument('--add-product', dest='add_product', action='append',
                                        choices=[x[1] for x in constants.MSP_PLANS],
                                        help='Allows MSP to use product. Can be repeated')
distributor_license_parser.add_argument('--remove-product', dest='remove_product', action='append',
                                        choices=[x[1] for x in constants.MSP_PLANS],
                                        help='Disables MSP to use product. Can be repeated')
distributor_license_parser.add_argument('--add-addon', dest='add_addon', action='append',
                                        choices=[x[0] for x in constants.MSP_ADDONS],
                                        help='Allows MSP to use add-on. Can be repeated')
distributor_license_parser.add_argument('--remove-addon', dest='remove_addon', action='append',
                                        choices=[x[0] for x in constants.MSP_ADDONS],
                                        help='Disables MSP to use add-on. Can be repeated')
distributor_license_parser.add_argument('--max-file-plan', dest='max_file_plan', action='store',
                                        choices=[x[2] for x in constants.MSP_FILE_PLANS],
                                        help='Maximum available file plan.')
distributor_license_parser.add_argument('--allocate-unlimited', dest='allocate_unlimited', action='store', choices=['on', 'off'],
                                        help='Allow MSPs to allocate unlimited licenses.')
distributor_license_parser.add_argument('msp', action='store', metavar='MSP_NAME', nargs='*',
                                        help='Managed Company Provider (name, ID, or name pattern).')


class DistributorCommand(GroupCommand):
    def __init__(self):
        super(DistributorCommand, self).__init__()
        self.register_command('info', DistributorInfoCommand(), description='Lists information about MSPs', alias='i')
        self.register_command('license', DistributorLicenseCommand(), alias='l',
                              description='Lists or changes available products/addons/file plans for named MSP (name or ID)')
        self.register_command('msp-info', DistributorMspInfoCommand(), alias='mi',
                              description='Lists MSP info for named MSP (name or ID)')

        self.default_verb = 'info'


class DistributorMixin:
    @staticmethod
    def convert_add_ons(add_on):  # type: (enterprise_pb2.LicenseAddOn) -> dict
        return {
            'name': add_on.name,
            'enabled': add_on.enabled,
            'is_trial': add_on.isTrial,
            'expiration': add_on.expiration,
            'created': add_on.created,
            'seats': add_on.seats,
            'activation_time': add_on.activationTime,
            'included_in_product': add_on.includedInProduct,
            'api_call_count': add_on.apiCallCount
        }

    @staticmethod
    def get_distributor_list(params, reload=False):
        # type: (KeeperParams, Optional[bool]) -> Optional[list]
        if not params.enterprise:
            return

        if reload is True:
            if 'distributors' in params.enterprise:
                del params.enterprise['distributors']
        if 'distributors' not in params.enterprise:
            rs = api.communicate_rest(params, None, 'distributor/get_distributor_info',
                                      rs_type=enterprise_pb2.GetDistributorInfoResponse)
            msps = []
            for d in rs.distributors:
                distributor_name = d.name
                for p_msp in d.mspInfos:
                    msp = {
                        'distributor_name': distributor_name,
                        'enterprise_id': p_msp.enterpriseId,
                        'enterprise_name': p_msp.enterpriseName,
                        'allocated_licenses': p_msp.allocatedLicenses,
                        'allowed_mc_products': list(p_msp.allowedMcProducts),
                        'allowed_add_ons': list(p_msp.allowedAddOns),
                        'max_file_plan_type': p_msp.maxFilePlanType,
                        'managed_companies': [],
                        'allow_unlimited_licenses': p_msp.allowUnlimitedLicenses,
                        'add_ons': [DistributorMixin.convert_add_ons(x) for x in p_msp.addOns]
                    }
                    for p_mc in p_msp.managedCompanies:
                        mc = {
                            'mc_enterprise_id': p_mc.mcEnterpriseId,
                            'mc_enterprise_name': p_mc.mcEnterpriseName,
                            'msp_node_id': p_mc.mspNodeId,
                            'number_of_seats': p_mc.numberOfSeats,
                            'number_of_users': p_mc.numberOfUsers,
                            'product_id': p_mc.productId,
                            'is_expired': p_mc.isExpired,
                            'file_plan_type': p_mc.filePlanType,
                            'add_ons': [DistributorMixin.convert_add_ons(x) for x in p_mc.addOns]
                        }
                        msp['managed_companies'].append(mc)
                    msps.append(msp)
            params.enterprise['distributors'] = msps

        return params.enterprise.get('distributors')

    @staticmethod
    def get_msp(params, msp_name, reload=False):    # type: (KeeperParams, str, bool) -> Optional[dict]
        msps = DistributorMixin.get_distributor_list(params, reload)
        msp_name_l = msp_name.lower() if isinstance(msp_name, str) else str(msp_name)
        return next((x for x in msps if str(x['enterprise_id']) == msp_name_l or x['enterprise_name'].lower() == msp_name_l), None)


class DistributorInfoCommand(EnterpriseCommand, DistributorMixin):
    def get_parser(self):
        return distributor_info_parser

    def execute(self, params, **kwargs):
        msps = DistributorMixin.get_distributor_list(params, kwargs.get('reload'))
        if not isinstance(msps, list):
            raise CommandError('', 'This command is only available for Distributors.')

        show_mc_details = kwargs.get('mc_details') is True
        output_format = kwargs.get('format', 'table')
        verbose = kwargs.get('verbose') is True
        if output_format == 'json':
            return json.dumps(msps, indent=4)
        else:
            right_align = (2, 6, 7) if show_mc_details else (2, 5)
            header = ['ID', 'MSP Name', '# MC\'s']
            if show_mc_details:
                header.extend(('MC Name', 'Plan', 'Storage', 'Total Licenses'))
            else:
                header.extend(('Unlimited allowed', 'Allowed Addons'))
            header.extend(['Licenses used', 'KSM', 'KCM', 'PAM', 'Addons'])
            addon_lookup = {x[0]: x[3] for x in constants.MSP_ADDONS}

            table = []
            for msp in msps:
                msp_name = msp.get('enterprise_name')
                if len(msp_name) > 40 and not verbose:
                    msp_name = msp_name[:37] + '...'
                row = [msp.get('enterprise_id'), msp_name, len(msp.get('managed_companies', []))]
                if show_mc_details:
                    row.extend((None, None, None, None))
                else:
                    unlimited_allowed = msp.get('allow_unlimited_licenses')
                    row.extend(('Allowed' if unlimited_allowed else 'Not Allowed',
                                [addon_lookup.get(x.lower(), x) for x in msp.get('allowed_add_ons', [])]))
                add_ons = msp.get('add_ons', [])
                kcm = next((x for x in add_ons if x and x['name'].lower() == 'connection_manager'), None)
                kcm_text = 'enabled' if kcm else 'disabled'
                if kcm:
                    seats = kcm.get("seats")
                    if isinstance(seats, int) and seats > 0:
                        kcm_text = f'{seats} seats'

                ksm = next((x for x in add_ons if x and x['name'].lower() == 'secrets_manager'), None)
                ksm_text = 'enabled' if ksm else 'disabled'
                if ksm:
                    apis = ksm.get("api_call_count")
                    if isinstance(apis, int) and apis > 0:
                        ksm_text = f'{apis} Calls/Month'

                pam = next((x for x in add_ons if x and x['name'].lower() == 'privileged_access_manager'), None)
                pam_text = 'enabled' if kcm else 'disabled'
                if pam:
                    seats = pam.get("seats")
                    if isinstance(seats, int) and seats > 0:
                        kcm_text = f'{seats} seats'

                add_ons = [x['name'] for x in add_ons
                           if x['name'].lower() not in {'connection_manager', 'secrets_manager', 'privileged_access_manager'} and
                           not x['included_in_product']]

                row.extend((msp.get('allocated_licenses'),
                            ksm_text, kcm_text, pam_text,
                            [addon_lookup.get(x.lower(), x) for x in add_ons]))
                table.append(row)

                if show_mc_details:
                    product_lookup = {x[1]: x[2] for x in constants.MSP_PLANS}
                    file_plan_lookup = {x[1]: x[2] for x in constants.MSP_FILE_PLANS}
                    for mc in msp.get('managed_companies', []):
                        mc_name = mc.get('mc_enterprise_name')
                        product = mc.get('product_id')
                        product = product_lookup.get(product, product)
                        file_plan = mc.get('file_plan_type')
                        file_plan = file_plan_lookup.get(file_plan, file_plan)
                        is_expired = mc.get('is_expired', False)
                        if len(mc_name) > 40 and not verbose:
                            mc_name = mc_name[:37] + '...'
                        total_licences = mc.get('number_of_seats')
                        if total_licences > 2000000000:
                            total_licences = 'Unlimited'
                        used = '-' if is_expired else mc.get('number_of_users')

                        add_ons = list(mc.get('add_ons'))
                        kcm = next((x for x in add_ons if x and x['name'].lower() == 'connection_manager'), None)
                        kcm_text = 'enabled' if kcm else 'disabled'
                        if kcm:
                            seats = kcm.get("seats")
                            if isinstance(seats, int) and seats > 0:
                                kcm_text = f'{seats} seats'
                        ksm = next((x for x in add_ons if x and x['name'].lower() == 'secrets_manager'), None)
                        ksm_text = 'enabled' if ksm else 'disabled'
                        if ksm:
                            apis = ksm.get("api_call_count")
                            if isinstance(apis, int) and apis > 0:
                                ksm_text = f'{apis} Calls/Month'
                        pam = next((x for x in add_ons if x and x['name'].lower() == 'privileged_access_manager'), None)
                        pam_text = 'enabled' if pam else 'disabled'
                        if pam:
                            seats = kcm.get("seats")
                            if isinstance(pam, int) and seats > 0:
                                pam_text = f'{seats} seats'
                        add_ons = [x['name'] for x in add_ons
                                   if x['name'].lower() not in {'connection_manager', 'secrets_manager', 'privileged_access_manager'} and
                                   not x['included_in_product']]
                        row = [mc.get('mc_enterprise_id'), msp_name, None,
                               mc_name, product, file_plan, total_licences,
                               used, ksm_text, kcm_text, pam_text,
                               [addon_lookup.get(x.lower(), x) for x in add_ons]]
                        table.append(row)

            return dump_report_data(table, header, fmt=output_format, filename=kwargs.get('output'), row_number=True,
                                    right_align=right_align)


class DistributorMspInfoCommand(EnterpriseCommand, DistributorMixin):
    def get_parser(self):
        return distributor_msp_info_parser

    def execute(self, params, **kwargs):
        msp_name = kwargs.get('msp')
        if not msp_name:
            raise CommandError('', '"msp" parameter is required.')
        msp = DistributorMixin.get_msp(params, msp_name)
        if not msp:
            raise CommandError('', f'MSP \"{msp_name}\" not found.')

        product_lookup = {x[1]: x[2] for x in constants.MSP_PLANS}
        file_plan_lookup = {x[1]: x[2] for x in constants.MSP_FILE_PLANS}
        addon_lookup = {x[0]: x[3] for x in constants.MSP_ADDONS}
        output_format = kwargs.get('format', 'table')
        verbose = kwargs.get('verbose') is True
        if output_format == 'json':
            return json.dumps(msp, indent=4)
        else:
            header = ['ID', 'MC Name', 'Node ID', 'Plan', 'Storage', 'Addons', 'Allocated', 'Active']
            table = []
            for mc in msp.get('managed_companies', []):
                mc_name = mc.get('mc_enterprise_name')
                if len(mc_name) > 40 and not verbose:
                    mc_name = mc_name[:37] + '...'
                msp_node_id = mc.get('msp_node_id')
                allowed = mc.get('number_of_seats')
                is_expired = mc.get('is_expired', False)
                if allowed > 2000000000:
                    allowed = 'Unlimited'
                used = '-' if is_expired else mc.get('number_of_users')
                add_ons = list(mc.get('add_ons'))
                if verbose:
                    ao = []
                    for x in add_ons:
                        addon_name = addon_lookup.get(x['name'], x['name'])
                        seats = x.get('seats', 0)
                        if seats > 0:
                            addon_name += f' ({seats})'
                        ao.append(addon_name)
                    ao.sort()
                    addons = '\n'.join(ao)
                else:
                    addons = len(add_ons)
                product = mc.get('product_id')
                product = product_lookup.get(product, product)
                file_plan = mc.get('file_plan_type')
                file_plan = file_plan_lookup.get(file_plan, file_plan)
                row = [mc.get('mc_enterprise_id'), mc_name, msp_node_id, product, file_plan, addons, allowed, used]

                table.append(row)

            right_align = (6, 7) if verbose else (5, 6, 7)
            return dump_report_data(table, header, fmt=output_format, filename=kwargs.get('output'), row_number=True, right_align=right_align)


class DistributorLicenseCommand(EnterpriseCommand, DistributorMixin):
    def get_parser(self):
        return distributor_license_parser

    def execute(self, params, **kwargs):
        msp_dict = {x['enterprise_id']: x for x in  DistributorMixin.get_distributor_list(params, False)}
        msps = set()  # type: Set[int]

        msp_names = kwargs.get('msp')
        if msp_names:
            if isinstance(msp_names, (str, int)):
                msp_names = [msp_names]

            for msp_name in msp_names:
                if isinstance(msp_name, int) or isinstance(msp_name, str) and msp_name.isnumeric():
                    try:
                        msp_id = int(msp_name)
                        if msp_id in msp_dict:
                            msps.add(msp_id)
                    except:
                        pass
                if isinstance(msp_name, str):
                    pattern = fnmatch.translate(msp_name)
                    for msp_id, msp in msp_dict.items():
                        if re.match(pattern, msp['enterprise_name'], re.IGNORECASE):
                            msps.add(msp_id)
        else:
            edit_opts = {'add_product', 'remove_product', 'add_addon', 'remove_addon', 'max_file_plan', 'allocate_unlimited'}
            if edit_opts.intersection((key for key, value in kwargs.items() if value)):
                logging.warning('Changes to MSP configuration will be ignored.\nPass MSP ID, name, or name search pattern.')

        if len(msps) > 0:
            for msp_id in msps:
                msp = msp_dict[msp_id]

                has_updates = False
                products = set((x.lower() for x in msp['allowed_mc_products']))
                add_products = kwargs.get('add_product')
                if isinstance(add_products, str):
                    add_products = [add_products]
                remove_products = kwargs.get('remove_product')
                if isinstance(remove_products, str):
                    remove_products = [remove_products]

                if add_products or remove_products:
                    has_updates = True
                    all_products = {x[1].lower() for x in constants.MSP_PLANS}
                    if isinstance(add_products, (list, tuple)):
                        for product in add_products:
                            product = product.lower()
                            if product not in all_products:
                                raise CommandError('', f'Unknown Product: {product}')
                            products.add(product)

                    if isinstance(remove_products, (list, tuple)):
                        for product in remove_products:
                            product = product.lower()
                            if product not in all_products:
                                raise CommandError('', f'Unknown Product: {product}')
                            if product in products:
                                products.remove(product)

                addons = set((x.lower() for x in msp['allowed_add_ons']))
                add_addons = kwargs.get('add_addon')
                if isinstance(add_addons, str):
                    add_addons = [add_addons]
                remove_addons = kwargs.get('remove_addon')
                if isinstance(remove_addons, str):
                    remove_addons = [remove_addons]

                if add_addons or remove_addons:
                    has_updates = True
                    all_addons = {x[0].lower() for x in constants.MSP_ADDONS}
                    if isinstance(add_addons, (list, tuple)):
                        for addon in add_addons:
                            if addon.lower() not in all_addons:
                                raise CommandError('', f'Unknown Add On: {addon}')
                            addons.add(addon.lower())

                    if isinstance(remove_addons, (list, tuple)):
                        for addon in remove_addons:
                            addon = addon.lower()
                            if addon not in all_addons:
                                raise CommandError('', f'Unknown product: {addon}')
                            if addon in addons:
                                addons.remove(addon)

                max_file_plan = msp['max_file_plan_type']
                file_plan = kwargs.get('max_file_plan')
                if file_plan and isinstance(file_plan, str):
                    has_updates = True
                    all_plans = {x[2].lower(): x[1] for x in constants.MSP_FILE_PLANS}
                    l_file_plan = file_plan.lower()
                    if l_file_plan not in all_plans:
                        raise CommandError('', f'Unknown File Plan: {file_plan}')
                    max_file_plan = all_plans[l_file_plan]

                allow_unlimited = msp['allow_unlimited_licenses']  # type: bool
                allocate_unlimited = kwargs.get('allocate_unlimited')
                if allocate_unlimited:
                    if allocate_unlimited == 'on':
                        allow_unlimited = True
                    else:
                        allow_unlimited = False
                    if msp.get('allow_unlimited_licenses') != allow_unlimited:
                        has_updates = True

                if has_updates:
                    if len(msps) > 1:
                        logging.info('Updating %s', msp['enterprise_name'])
                    rq = enterprise_pb2.UpdateMSPPermitsRequest()
                    rq.mspEnterpriseId = msp['enterprise_id']
                    rq.allowUnlimitedLicenses = allow_unlimited
                    all_products = {x[1].lower(): x[1] for x in constants.MSP_PLANS}
                    rq.allowedMcProducts.extend(all_products.get(x, x) for x in products)
                    all_addons = {x[0].lower(): x[0] for x in constants.MSP_ADDONS}
                    rq.allowedAddOns.extend(all_addons.get(x, x) for x in addons)
                    rq.maxFilePlanType = max_file_plan

                    api.communicate_rest(params, rq, 'distributor/update_msp_permits')
            msp_dict = {x['enterprise_id']: x for x in  DistributorMixin.get_distributor_list(params, True)}
        else:
            if msp_names:
                logging.warning('MSP \"%s\" not found.', ', '.join(msp_names))

        # display
        table = []
        header = ['MSP ID', 'MSP Name', 'Unlimited Allowed', 'Allowed Products', 'Allowed Addons', 'Max File Plan']
        for msp_id, msp in msp_dict.items():
            if len(msps) > 0 and msp_id not in msps:
                continue

            row = [msp['enterprise_id'], msp['enterprise_name'], msp['allow_unlimited_licenses']]
            products = msp['allowed_mc_products']
            if products and isinstance(products, list):
                all_products = {x[1].lower(): x[2] for x in constants.MSP_PLANS}
                products = [x + f' ({all_products.get(x.lower(), "")})' for x in products]
            else:
                products = None
            row.append(products)

            addons = msp['allowed_add_ons']
            if addons and isinstance(addons, list):
                all_addons = {x[0].lower(): x[3] for x in constants.MSP_ADDONS}
                addons = [x.lower() + f' ({all_addons.get(x.lower(), "")})' for x in addons]
            else:
                addons = None
            row.append(addons)

            max_file_plan = msp['max_file_plan_type']
            all_file_plans = {x[1].lower(): x[2] for x in constants.MSP_FILE_PLANS}
            row.append(all_file_plans.get(max_file_plan.lower(), max_file_plan))
            table.append(row)
        dump_report_data(table, headers=header, right_align=(0,))



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import io, html
from pathlib import Path
import sys
import json
import logging
from typing import Any, Tuple, Optional
from .config_reader import ConfigReader
from .exceptions import CommandExecutionError
from .parse_keeper_response import parse_keeper_response
from ..core.globals import get_current_params
from ..decorators.logging import logger, debug_decorator, sanitize_debug_data
from ... import cli, utils
from ...__main__ import get_params_from_config
from ...service.config.service_config import ServiceConfig
from ...crypto import encrypt_aes_v2

class CommandExecutor:
    @staticmethod
    @debug_decorator
    def validate_command(command: str) -> Optional[Tuple[dict, int]]:
        if not command:
            return {"error": "No command provided."}, 400
        return None

    @staticmethod
    @debug_decorator
    def validate_session() -> Optional[Tuple[dict, int]]:
        params = get_current_params()
        if not params:
            return {"error": "No active session. Please log in through the CLI first."}, 401
        return None

    @staticmethod
    @debug_decorator
    def capture_output_and_logs(params: Any, command: str) -> Tuple[Any, str, str]:
        """Capture both stdout/stderr and logging output from command execution."""
        captured_stdout = io.StringIO()
        captured_stderr = io.StringIO()
        captured_logs = io.StringIO()
        
        # Create a temporary log handler to capture command logs
        temp_handler = logging.StreamHandler(captured_logs)
        temp_handler.setFormatter(logging.Formatter('%(message)s'))
        
        # Get the root logger to capture all logging output
        root_logger = logging.getLogger()
        original_level = root_logger.level
        original_handlers = root_logger.handlers[:]
        
        original_stdout = sys.stdout
        original_stderr = sys.stderr
        
        sys.stdout = captured_stdout
        sys.stderr = captured_stderr
        
        try:
            # Add our temporary handler and set appropriate level
            root_logger.addHandler(temp_handler)
            root_logger.setLevel(logging.INFO)
            
            return_value = cli.do_command(params, command)
            
            stdout_content = captured_stdout.getvalue()
            stderr_content = captured_stderr.getvalue()
            log_content = captured_logs.getvalue()
            
            # Combine stdout and stderr
            stdout_clean = stdout_content.strip()
            stderr_clean = stderr_content.strip()
            log_clean = log_content.strip()
            
            if stderr_clean and stdout_clean:
                combined_output = stderr_clean + '\n' + stdout_clean
            else:
                combined_output = stderr_clean or stdout_clean
            
            return return_value, combined_output, log_clean
        except Exception as e:
            # If there's an exception, capture any error output
            stderr_content = captured_stderr.getvalue()
            log_content = captured_logs.getvalue()
            error_output = stderr_content.strip()
            
            if error_output:
                raise CommandExecutionError(f"Command failed: {error_output}")
            raise
        finally:
            sys.stdout = original_stdout
            sys.stderr = original_stderr
            # Restore original logging configuration
            root_logger.removeHandler(temp_handler)
            root_logger.setLevel(original_level)
            temp_handler.close()


    @staticmethod
    @debug_decorator
    def encrypt_response(response: Any) -> bytes:
        encryption_key = ConfigReader.read_config('encryption_private_key')
        if encryption_key:
            try:
                encryption_key_bytes = encryption_key.encode('utf-8')
                response_bytes = json.dumps(response).encode('utf-8')
                return encrypt_aes_v2(response_bytes, encryption_key_bytes)
            except Exception as e:
                raise
        return response

    @classmethod
    def execute(cls, command: str) -> Tuple[Any, int]:
        logger.debug(f"Executing command: {command}")
        
        validation_error = cls.validate_command(command)
        if validation_error:
            return validation_error
        
        from ..core.globals import ensure_params_loaded
        try:
            params = ensure_params_loaded()
            # Set service mode flag to bypass master password enforcement
            if params:
                params.service_mode = True

            command = html.unescape(command)
            return_value, printed_output, log_output = CommandExecutor.capture_output_and_logs(params, command)
            response = return_value if return_value else printed_output

            # Debug logging with sanitization
            sanitized_output = sanitize_debug_data(printed_output)
            sanitized_logs = sanitize_debug_data(log_output)
            logger.debug(f"After capture_output - return_value: '{return_value}', printed_output: '{sanitized_output}', log_output: '{sanitized_logs}'")
            sanitized_response = sanitize_debug_data(str(response))
            logger.debug(f"Final response: '{sanitized_response}', response type: {type(response)}")

            cli.do_command(params, 'sync-down')
            
            # Always let the parser handle the response (including empty responses and logs)
            response = parse_keeper_response(command, response, log_output)
            
            response = CommandExecutor.encrypt_response(response)
            logger.debug("Command executed successfully")
            return response, 200
        except CommandExecutionError as e:
            # Return the actual command error instead of generic "server busy"
            logger.error(f"Command execution error: {e}")
            error_response = {
                "status": "error",
                "error": str(e)
            }
            return error_response, 400
        except Exception as e:
            # Log unexpected errors and return a proper error response
            logger.error(f"Unexpected error during command execution: {e}")
            error_response = {
                "status": "error",
                "error": f"Unexpected error: {str(e)}"
            }
            return error_response, 500


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import os
import sys
from typing import Optional
from ..decorators.logging import debug_decorator

class TerminalHandler:
    @staticmethod
    @debug_decorator
    def get_terminal_info() -> Optional[str]:
        """Get terminal information based on platform."""
        try:
            if sys.platform.startswith('win'):
                return os.environ.get('COMSPEC')
            return os.ttyname(sys.stdout.fileno()) if sys.stdout.isatty() else None
        except Exception as e:
            return None

    @staticmethod
    def notify_other_terminal(terminal: str) -> None:
        """Notify other terminal about service stop."""
        if not sys.platform.startswith('win'):
            try:
                with open(terminal, 'w') as term:
                    term.write("Commander Service stopped by another terminal\n")
            except Exception as e:
                pass


import record_pb2 as _record_pb2
import breachwatch_pb2 as _breachwatch_pb2
import APIRequest_pb2 as _APIRequest_pb2
import enterprise_pb2 as _enterprise_pb2
import NotificationCenter_pb2 as _NotificationCenter_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CacheStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    KEEP: _ClassVar[CacheStatus]
    CLEAR: _ClassVar[CacheStatus]

class RecordRotationStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    RRST_NOT_ROTATED: _ClassVar[RecordRotationStatus]
    RRST_IN_PROGRESS: _ClassVar[RecordRotationStatus]
    RRST_SUCCESS: _ClassVar[RecordRotationStatus]
    RRST_FAILURE: _ClassVar[RecordRotationStatus]
KEEP: CacheStatus
CLEAR: CacheStatus
RRST_NOT_ROTATED: RecordRotationStatus
RRST_IN_PROGRESS: RecordRotationStatus
RRST_SUCCESS: RecordRotationStatus
RRST_FAILURE: RecordRotationStatus

class SyncDownRequest(_message.Message):
    __slots__ = ["continuationToken", "dataVersion"]
    CONTINUATIONTOKEN_FIELD_NUMBER: _ClassVar[int]
    DATAVERSION_FIELD_NUMBER: _ClassVar[int]
    continuationToken: bytes
    dataVersion: int
    def __init__(self, continuationToken: _Optional[bytes] = ..., dataVersion: _Optional[int] = ...) -> None: ...

class SyncDownResponse(_message.Message):
    __slots__ = ["continuationToken", "hasMore", "cacheStatus", "userFolders", "sharedFolders", "userFolderSharedFolders", "sharedFolderFolders", "records", "recordMetaData", "nonSharedData", "recordLinks", "userFolderRecords", "sharedFolderRecords", "sharedFolderFolderRecords", "sharedFolderUsers", "sharedFolderTeams", "recordAddAuditData", "teams", "sharingChanges", "profile", "profilePic", "pendingTeamMembers", "breachWatchRecords", "userAuths", "breachWatchSecurityData", "reusedPasswords", "removedUserFolders", "removedSharedFolders", "removedUserFolderSharedFolders", "removedSharedFolderFolders", "removedRecords", "removedRecordLinks", "removedUserFolderRecords", "removedSharedFolderRecords", "removedSharedFolderFolderRecords", "removedSharedFolderUsers", "removedSharedFolderTeams", "removedTeams", "ksmAppShares", "ksmAppClients", "shareInvitations", "diagnostics", "recordRotations", "users", "removedUsers", "securityScoreData", "notificationSync"]
    CONTINUATIONTOKEN_FIELD_NUMBER: _ClassVar[int]
    HASMORE_FIELD_NUMBER: _ClassVar[int]
    CACHESTATUS_FIELD_NUMBER: _ClassVar[int]
    USERFOLDERS_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERS_FIELD_NUMBER: _ClassVar[int]
    USERFOLDERSHAREDFOLDERS_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERFOLDERS_FIELD_NUMBER: _ClassVar[int]
    RECORDS_FIELD_NUMBER: _ClassVar[int]
    RECORDMETADATA_FIELD_NUMBER: _ClassVar[int]
    NONSHAREDDATA_FIELD_NUMBER: _ClassVar[int]
    RECORDLINKS_FIELD_NUMBER: _ClassVar[int]
    USERFOLDERRECORDS_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERRECORDS_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERFOLDERRECORDS_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERUSERS_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERTEAMS_FIELD_NUMBER: _ClassVar[int]
    RECORDADDAUDITDATA_FIELD_NUMBER: _ClassVar[int]
    TEAMS_FIELD_NUMBER: _ClassVar[int]
    SHARINGCHANGES_FIELD_NUMBER: _ClassVar[int]
    PROFILE_FIELD_NUMBER: _ClassVar[int]
    PROFILEPIC_FIELD_NUMBER: _ClassVar[int]
    PENDINGTEAMMEMBERS_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHRECORDS_FIELD_NUMBER: _ClassVar[int]
    USERAUTHS_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHSECURITYDATA_FIELD_NUMBER: _ClassVar[int]
    REUSEDPASSWORDS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDUSERFOLDERS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDSHAREDFOLDERS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDUSERFOLDERSHAREDFOLDERS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDSHAREDFOLDERFOLDERS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDRECORDS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDRECORDLINKS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDUSERFOLDERRECORDS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDSHAREDFOLDERRECORDS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDSHAREDFOLDERFOLDERRECORDS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDSHAREDFOLDERUSERS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDSHAREDFOLDERTEAMS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDTEAMS_FIELD_NUMBER: _ClassVar[int]
    KSMAPPSHARES_FIELD_NUMBER: _ClassVar[int]
    KSMAPPCLIENTS_FIELD_NUMBER: _ClassVar[int]
    SHAREINVITATIONS_FIELD_NUMBER: _ClassVar[int]
    DIAGNOSTICS_FIELD_NUMBER: _ClassVar[int]
    RECORDROTATIONS_FIELD_NUMBER: _ClassVar[int]
    USERS_FIELD_NUMBER: _ClassVar[int]
    REMOVEDUSERS_FIELD_NUMBER: _ClassVar[int]
    SECURITYSCOREDATA_FIELD_NUMBER: _ClassVar[int]
    NOTIFICATIONSYNC_FIELD_NUMBER: _ClassVar[int]
    continuationToken: bytes
    hasMore: bool
    cacheStatus: CacheStatus
    userFolders: _containers.RepeatedCompositeFieldContainer[UserFolder]
    sharedFolders: _containers.RepeatedCompositeFieldContainer[SharedFolder]
    userFolderSharedFolders: _containers.RepeatedCompositeFieldContainer[UserFolderSharedFolder]
    sharedFolderFolders: _containers.RepeatedCompositeFieldContainer[SharedFolderFolder]
    records: _containers.RepeatedCompositeFieldContainer[Record]
    recordMetaData: _containers.RepeatedCompositeFieldContainer[RecordMetaData]
    nonSharedData: _containers.RepeatedCompositeFieldContainer[NonSharedData]
    recordLinks: _containers.RepeatedCompositeFieldContainer[RecordLink]
    userFolderRecords: _containers.RepeatedCompositeFieldContainer[UserFolderRecord]
    sharedFolderRecords: _containers.RepeatedCompositeFieldContainer[SharedFolderRecord]
    sharedFolderFolderRecords: _containers.RepeatedCompositeFieldContainer[SharedFolderFolderRecord]
    sharedFolderUsers: _containers.RepeatedCompositeFieldContainer[SharedFolderUser]
    sharedFolderTeams: _containers.RepeatedCompositeFieldContainer[SharedFolderTeam]
    recordAddAuditData: _containers.RepeatedScalarFieldContainer[bytes]
    teams: _containers.RepeatedCompositeFieldContainer[Team]
    sharingChanges: _containers.RepeatedCompositeFieldContainer[SharingChange]
    profile: Profile
    profilePic: ProfilePic
    pendingTeamMembers: _containers.RepeatedCompositeFieldContainer[PendingTeamMember]
    breachWatchRecords: _containers.RepeatedCompositeFieldContainer[BreachWatchRecord]
    userAuths: _containers.RepeatedCompositeFieldContainer[UserAuth]
    breachWatchSecurityData: _containers.RepeatedCompositeFieldContainer[BreachWatchSecurityData]
    reusedPasswords: ReusedPasswords
    removedUserFolders: _containers.RepeatedScalarFieldContainer[bytes]
    removedSharedFolders: _containers.RepeatedScalarFieldContainer[bytes]
    removedUserFolderSharedFolders: _containers.RepeatedCompositeFieldContainer[UserFolderSharedFolder]
    removedSharedFolderFolders: _containers.RepeatedCompositeFieldContainer[SharedFolderFolder]
    removedRecords: _containers.RepeatedScalarFieldContainer[bytes]
    removedRecordLinks: _containers.RepeatedCompositeFieldContainer[RecordLink]
    removedUserFolderRecords: _containers.RepeatedCompositeFieldContainer[UserFolderRecord]
    removedSharedFolderRecords: _containers.RepeatedCompositeFieldContainer[SharedFolderRecord]
    removedSharedFolderFolderRecords: _containers.RepeatedCompositeFieldContainer[SharedFolderFolderRecord]
    removedSharedFolderUsers: _containers.RepeatedCompositeFieldContainer[SharedFolderUser]
    removedSharedFolderTeams: _containers.RepeatedCompositeFieldContainer[SharedFolderTeam]
    removedTeams: _containers.RepeatedScalarFieldContainer[bytes]
    ksmAppShares: _containers.RepeatedCompositeFieldContainer[KsmChange]
    ksmAppClients: _containers.RepeatedCompositeFieldContainer[KsmChange]
    shareInvitations: _containers.RepeatedCompositeFieldContainer[ShareInvitation]
    diagnostics: SyncDiagnostics
    recordRotations: _containers.RepeatedCompositeFieldContainer[RecordRotation]
    users: _containers.RepeatedCompositeFieldContainer[User]
    removedUsers: _containers.RepeatedScalarFieldContainer[bytes]
    securityScoreData: _containers.RepeatedCompositeFieldContainer[SecurityScoreData]
    notificationSync: _containers.RepeatedCompositeFieldContainer[_NotificationCenter_pb2.NotificationWrapper]
    def __init__(self, continuationToken: _Optional[bytes] = ..., hasMore: bool = ..., cacheStatus: _Optional[_Union[CacheStatus, str]] = ..., userFolders: _Optional[_Iterable[_Union[UserFolder, _Mapping]]] = ..., sharedFolders: _Optional[_Iterable[_Union[SharedFolder, _Mapping]]] = ..., userFolderSharedFolders: _Optional[_Iterable[_Union[UserFolderSharedFolder, _Mapping]]] = ..., sharedFolderFolders: _Optional[_Iterable[_Union[SharedFolderFolder, _Mapping]]] = ..., records: _Optional[_Iterable[_Union[Record, _Mapping]]] = ..., recordMetaData: _Optional[_Iterable[_Union[RecordMetaData, _Mapping]]] = ..., nonSharedData: _Optional[_Iterable[_Union[NonSharedData, _Mapping]]] = ..., recordLinks: _Optional[_Iterable[_Union[RecordLink, _Mapping]]] = ..., userFolderRecords: _Optional[_Iterable[_Union[UserFolderRecord, _Mapping]]] = ..., sharedFolderRecords: _Optional[_Iterable[_Union[SharedFolderRecord, _Mapping]]] = ..., sharedFolderFolderRecords: _Optional[_Iterable[_Union[SharedFolderFolderRecord, _Mapping]]] = ..., sharedFolderUsers: _Optional[_Iterable[_Union[SharedFolderUser, _Mapping]]] = ..., sharedFolderTeams: _Optional[_Iterable[_Union[SharedFolderTeam, _Mapping]]] = ..., recordAddAuditData: _Optional[_Iterable[bytes]] = ..., teams: _Optional[_Iterable[_Union[Team, _Mapping]]] = ..., sharingChanges: _Optional[_Iterable[_Union[SharingChange, _Mapping]]] = ..., profile: _Optional[_Union[Profile, _Mapping]] = ..., profilePic: _Optional[_Union[ProfilePic, _Mapping]] = ..., pendingTeamMembers: _Optional[_Iterable[_Union[PendingTeamMember, _Mapping]]] = ..., breachWatchRecords: _Optional[_Iterable[_Union[BreachWatchRecord, _Mapping]]] = ..., userAuths: _Optional[_Iterable[_Union[UserAuth, _Mapping]]] = ..., breachWatchSecurityData: _Optional[_Iterable[_Union[BreachWatchSecurityData, _Mapping]]] = ..., reusedPasswords: _Optional[_Union[ReusedPasswords, _Mapping]] = ..., removedUserFolders: _Optional[_Iterable[bytes]] = ..., removedSharedFolders: _Optional[_Iterable[bytes]] = ..., removedUserFolderSharedFolders: _Optional[_Iterable[_Union[UserFolderSharedFolder, _Mapping]]] = ..., removedSharedFolderFolders: _Optional[_Iterable[_Union[SharedFolderFolder, _Mapping]]] = ..., removedRecords: _Optional[_Iterable[bytes]] = ..., removedRecordLinks: _Optional[_Iterable[_Union[RecordLink, _Mapping]]] = ..., removedUserFolderRecords: _Optional[_Iterable[_Union[UserFolderRecord, _Mapping]]] = ..., removedSharedFolderRecords: _Optional[_Iterable[_Union[SharedFolderRecord, _Mapping]]] = ..., removedSharedFolderFolderRecords: _Optional[_Iterable[_Union[SharedFolderFolderRecord, _Mapping]]] = ..., removedSharedFolderUsers: _Optional[_Iterable[_Union[SharedFolderUser, _Mapping]]] = ..., removedSharedFolderTeams: _Optional[_Iterable[_Union[SharedFolderTeam, _Mapping]]] = ..., removedTeams: _Optional[_Iterable[bytes]] = ..., ksmAppShares: _Optional[_Iterable[_Union[KsmChange, _Mapping]]] = ..., ksmAppClients: _Optional[_Iterable[_Union[KsmChange, _Mapping]]] = ..., shareInvitations: _Optional[_Iterable[_Union[ShareInvitation, _Mapping]]] = ..., diagnostics: _Optional[_Union[SyncDiagnostics, _Mapping]] = ..., recordRotations: _Optional[_Iterable[_Union[RecordRotation, _Mapping]]] = ..., users: _Optional[_Iterable[_Union[User, _Mapping]]] = ..., removedUsers: _Optional[_Iterable[bytes]] = ..., securityScoreData: _Optional[_Iterable[_Union[SecurityScoreData, _Mapping]]] = ..., notificationSync: _Optional[_Iterable[_Union[_NotificationCenter_pb2.NotificationWrapper, _Mapping]]] = ...) -> None: ...

class UserFolder(_message.Message):
    __slots__ = ["folderUid", "parentUid", "userFolderKey", "keyType", "revision", "data"]
    FOLDERUID_FIELD_NUMBER: _ClassVar[int]
    PARENTUID_FIELD_NUMBER: _ClassVar[int]
    USERFOLDERKEY_FIELD_NUMBER: _ClassVar[int]
    KEYTYPE_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    folderUid: bytes
    parentUid: bytes
    userFolderKey: bytes
    keyType: _record_pb2.RecordKeyType
    revision: int
    data: bytes
    def __init__(self, folderUid: _Optional[bytes] = ..., parentUid: _Optional[bytes] = ..., userFolderKey: _Optional[bytes] = ..., keyType: _Optional[_Union[_record_pb2.RecordKeyType, str]] = ..., revision: _Optional[int] = ..., data: _Optional[bytes] = ...) -> None: ...

class SharedFolder(_message.Message):
    __slots__ = ["sharedFolderUid", "revision", "sharedFolderKey", "keyType", "data", "defaultManageRecords", "defaultManageUsers", "defaultCanEdit", "defaultCanReshare", "cacheStatus", "owner", "ownerAccountUid", "name"]
    SHAREDFOLDERUID_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERKEY_FIELD_NUMBER: _ClassVar[int]
    KEYTYPE_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    DEFAULTMANAGERECORDS_FIELD_NUMBER: _ClassVar[int]
    DEFAULTMANAGEUSERS_FIELD_NUMBER: _ClassVar[int]
    DEFAULTCANEDIT_FIELD_NUMBER: _ClassVar[int]
    DEFAULTCANRESHARE_FIELD_NUMBER: _ClassVar[int]
    CACHESTATUS_FIELD_NUMBER: _ClassVar[int]
    OWNER_FIELD_NUMBER: _ClassVar[int]
    OWNERACCOUNTUID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    sharedFolderUid: bytes
    revision: int
    sharedFolderKey: bytes
    keyType: _record_pb2.RecordKeyType
    data: bytes
    defaultManageRecords: bool
    defaultManageUsers: bool
    defaultCanEdit: bool
    defaultCanReshare: bool
    cacheStatus: CacheStatus
    owner: str
    ownerAccountUid: bytes
    name: bytes
    def __init__(self, sharedFolderUid: _Optional[bytes] = ..., revision: _Optional[int] = ..., sharedFolderKey: _Optional[bytes] = ..., keyType: _Optional[_Union[_record_pb2.RecordKeyType, str]] = ..., data: _Optional[bytes] = ..., defaultManageRecords: bool = ..., defaultManageUsers: bool = ..., defaultCanEdit: bool = ..., defaultCanReshare: bool = ..., cacheStatus: _Optional[_Union[CacheStatus, str]] = ..., owner: _Optional[str] = ..., ownerAccountUid: _Optional[bytes] = ..., name: _Optional[bytes] = ...) -> None: ...

class UserFolderSharedFolder(_message.Message):
    __slots__ = ["folderUid", "sharedFolderUid", "revision"]
    FOLDERUID_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERUID_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    folderUid: bytes
    sharedFolderUid: bytes
    revision: int
    def __init__(self, folderUid: _Optional[bytes] = ..., sharedFolderUid: _Optional[bytes] = ..., revision: _Optional[int] = ...) -> None: ...

class SharedFolderFolder(_message.Message):
    __slots__ = ["sharedFolderUid", "folderUid", "parentUid", "sharedFolderFolderKey", "keyType", "revision", "data"]
    SHAREDFOLDERUID_FIELD_NUMBER: _ClassVar[int]
    FOLDERUID_FIELD_NUMBER: _ClassVar[int]
    PARENTUID_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERFOLDERKEY_FIELD_NUMBER: _ClassVar[int]
    KEYTYPE_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    sharedFolderUid: bytes
    folderUid: bytes
    parentUid: bytes
    sharedFolderFolderKey: bytes
    keyType: _record_pb2.RecordKeyType
    revision: int
    data: bytes
    def __init__(self, sharedFolderUid: _Optional[bytes] = ..., folderUid: _Optional[bytes] = ..., parentUid: _Optional[bytes] = ..., sharedFolderFolderKey: _Optional[bytes] = ..., keyType: _Optional[_Union[_record_pb2.RecordKeyType, str]] = ..., revision: _Optional[int] = ..., data: _Optional[bytes] = ...) -> None: ...

class SharedFolderKey(_message.Message):
    __slots__ = ["sharedFolderUid", "sharedFolderKey", "keyType"]
    SHAREDFOLDERUID_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERKEY_FIELD_NUMBER: _ClassVar[int]
    KEYTYPE_FIELD_NUMBER: _ClassVar[int]
    sharedFolderUid: bytes
    sharedFolderKey: bytes
    keyType: _record_pb2.RecordKeyType
    def __init__(self, sharedFolderUid: _Optional[bytes] = ..., sharedFolderKey: _Optional[bytes] = ..., keyType: _Optional[_Union[_record_pb2.RecordKeyType, str]] = ...) -> None: ...

class Team(_message.Message):
    __slots__ = ["teamUid", "name", "teamKey", "teamKeyType", "teamPrivateKey", "restrictEdit", "restrictShare", "restrictView", "removedSharedFolders", "sharedFolderKeys", "teamEccPrivateKey", "teamEccPublicKey"]
    TEAMUID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    TEAMKEY_FIELD_NUMBER: _ClassVar[int]
    TEAMKEYTYPE_FIELD_NUMBER: _ClassVar[int]
    TEAMPRIVATEKEY_FIELD_NUMBER: _ClassVar[int]
    RESTRICTEDIT_FIELD_NUMBER: _ClassVar[int]
    RESTRICTSHARE_FIELD_NUMBER: _ClassVar[int]
    RESTRICTVIEW_FIELD_NUMBER: _ClassVar[int]
    REMOVEDSHAREDFOLDERS_FIELD_NUMBER: _ClassVar[int]
    SHAREDFOLDERKEYS_FIELD_NUMBER: _ClassVar[int]
    TEAMECCPRIVATEKEY_FIELD_NUMBER: _ClassVar[int]
    TEAMECCPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    teamUid: bytes
    name: str
    teamKey: bytes
    teamKeyType: _record_pb2.RecordKeyType
    teamPrivateKey: bytes
    restrictEdit: bool
    restrictShare: bool
    restrictView: bool
    removedSharedFolders: _containers.RepeatedScalarFieldContainer[bytes]
    sharedFolderKeys: _containers.RepeatedCompositeFieldContainer[SharedFolderKey]
    teamEccPrivateKey: bytes
    teamEccPublicKey: bytes
    def __init__(self, teamUid: _Optional[bytes] = ..., name: _Optional[str] = ..., teamKey: _Optional[bytes] = ..., teamKeyType: _Optional[_Union[_record_pb2.RecordKeyType, str]] = ..., teamPrivateKey: _Optional[bytes] = ..., restrictEdit: bool = ..., restrictShare: bool = ..., restrictView: bool = ..., removedSharedFolders: _Optional[_Iterable[bytes]] = ..., sharedFolderKeys: _Optional[_Iterable[_Union[SharedFolderKey, _Mapping]]] = ..., teamEccPrivateKey: _Optional[bytes] = ..., teamEccPublicKey: _Optional[bytes] = ...) -> None: ...

class Record(_message.Message):
    __slots__ = ["recordUid", "revision", "version", "shared", "clientModifiedTime", "data", "extra", "udata", "fileSize", "thumbnailSize"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    SHARED_FIELD_NUMBER: _ClassVar[int]
    CLIENTMODIFIEDTIME_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    EXTRA_FIELD_NUMBER: _ClassVar[int]
    UDATA_FIELD_NUMBER: _ClassVar[int]
    FILESIZE_FIELD_NUMBER: _ClassVar[int]
    THUMBNAILSIZE_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    revision: int
    version: int
    shared: bool
    clientModifiedTime: int
    data: bytes
    extra: bytes
    udata: str
    fileSize: int
    thumbnailSize: int
    def __init__(self, recordUid: _Optional[bytes] = ..., revision: _Optional[int] = ..., version: _Optional[int] = ..., shared: bool = ..., clientModifiedTime: _Optional[int] = ..., data: _Optional[bytes] = ..., extra: _Optional[bytes] = ..., udata: _Optional[str] = ..., fileSize: _Optional[int] = ..., thumbnailSize: _Optional[int] = ...) -> None: ...

class RecordLink(_message.Message):
    __slots__ = ["parentRecordUid", "childRecordUid", "recordKey", "revision"]
    PARENTRECORDUID_FIELD_NUMBER: _ClassVar[int]
    CHILDRECORDUID_FIELD_NUMBER: _ClassVar[int]
    RECORDKEY_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    parentRecordUid: bytes
    childRecordUid: bytes
    recordKey: bytes
    revision: int
    def __init__(self, parentRecordUid: _Optional[bytes] = ..., childRecordUid: _Optional[bytes] = ..., recordKey: _Optional[bytes] = ..., revision: _Optional[int] = ...) -> None: ...

class UserFolderRecord(_message.Message):
    __slots__ = ["folderUid", "recordUid", "revision"]
    FOLDERUID_FIELD_NUMBER: _ClassVar[int]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    folderUid: bytes
    recordUid: bytes
    revision: int
    def __init__(self, folderUid: _Optional[bytes] = ..., recordUid: _Optional[bytes] = ..., revision: _Optional[int] = ...) -> None: ...

class SharedFolderFolderRecord(_message.Message):
    __slots__ = ["sharedFolderUid", "folderUid", "recordUid", "revision"]
    SHAREDFOLDERUID_FIELD_NUMBER: _ClassVar[int]
    FOLDERUID_FIELD_NUMBER: _ClassVar[int]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    sharedFolderUid: bytes
    folderUid: bytes
    recordUid: bytes
    revision: int
    def __init__(self, sharedFolderUid: _Optional[bytes] = ..., folderUid: _Optional[bytes] = ..., recordUid: _Optional[bytes] = ..., revision: _Optional[int] = ...) -> None: ...

class NonSharedData(_message.Message):
    __slots__ = ["recordUid", "data"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    data: bytes
    def __init__(self, recordUid: _Optional[bytes] = ..., data: _Optional[bytes] = ...) -> None: ...

class RecordMetaData(_message.Message):
    __slots__ = ["recordUid", "owner", "recordKey", "recordKeyType", "canShare", "canEdit", "ownerAccountUid", "expiration", "expirationNotificationType", "ownerUsername"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    OWNER_FIELD_NUMBER: _ClassVar[int]
    RECORDKEY_FIELD_NUMBER: _ClassVar[int]
    RECORDKEYTYPE_FIELD_NUMBER: _ClassVar[int]
    CANSHARE_FIELD_NUMBER: _ClassVar[int]
    CANEDIT_FIELD_NUMBER: _ClassVar[int]
    OWNERACCOUNTUID_FIELD_NUMBER: _ClassVar[int]
    EXPIRATION_FIELD_NUMBER: _ClassVar[int]
    EXPIRATIONNOTIFICATIONTYPE_FIELD_NUMBER: _ClassVar[int]
    OWNERUSERNAME_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    owner: bool
    recordKey: bytes
    recordKeyType: _record_pb2.RecordKeyType
    canShare: bool
    canEdit: bool
    ownerAccountUid: bytes
    expiration: int
    expirationNotificationType: _record_pb2.TimerNotificationType
    ownerUsername: str
    def __init__(self, recordUid: _Optional[bytes] = ..., owner: bool = ..., recordKey: _Optional[bytes] = ..., recordKeyType: _Optional[_Union[_record_pb2.RecordKeyType, str]] = ..., canShare: bool = ..., canEdit: bool = ..., ownerAccountUid: _Optional[bytes] = ..., expiration: _Optional[int] = ..., expirationNotificationType: _Optional[_Union[_record_pb2.TimerNotificationType, str]] = ..., ownerUsername: _Optional[str] = ...) -> None: ...

class SharingChange(_message.Message):
    __slots__ = ["recordUid", "shared"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    SHARED_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    shared: bool
    def __init__(self, recordUid: _Optional[bytes] = ..., shared: bool = ...) -> None: ...

class Profile(_message.Message):
    __slots__ = ["data", "profileName", "revision"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    PROFILENAME_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    data: bytes
    profileName: str
    revision: int
    def __init__(self, data: _Optional[bytes] = ..., profileName: _Optional[str] = ..., revision: _Optional[int] = ...) -> None: ...

class ProfilePic(_message.Message):
    __slots__ = ["url", "revision"]
    URL_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    url: str
    revision: int
    def __init__(self, url: _Optional[str] = ..., revision: _Optional[int] = ...) -> None: ...

class PendingTeamMember(_message.Message):
    __slots__ = ["enterpriseUserId", "userPublicKey", "teamUids", "userEccPublicKey"]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    USERPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    TEAMUIDS_FIELD_NUMBER: _ClassVar[int]
    USERECCPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    enterpriseUserId: int
    userPublicKey: bytes
    teamUids: _containers.RepeatedScalarFieldContainer[bytes]
    userEccPublicKey: bytes
    def __init__(self, enterpriseUserId: _Optional[int] = ..., userPublicKey: _Optional[bytes] = ..., teamUids: _Optional[_Iterable[bytes]] = ..., userEccPublicKey: _Optional[bytes] = ...) -> None: ...

class BreachWatchRecord(_message.Message):
    __slots__ = ["recordUid", "data", "type", "scannedBy", "revision", "scannedByAccountUid"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    SCANNEDBY_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    SCANNEDBYACCOUNTUID_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    data: bytes
    type: _breachwatch_pb2.BreachWatchInfoType
    scannedBy: str
    revision: int
    scannedByAccountUid: bytes
    def __init__(self, recordUid: _Optional[bytes] = ..., data: _Optional[bytes] = ..., type: _Optional[_Union[_breachwatch_pb2.BreachWatchInfoType, str]] = ..., scannedBy: _Optional[str] = ..., revision: _Optional[int] = ..., scannedByAccountUid: _Optional[bytes] = ...) -> None: ...

class UserAuth(_message.Message):
    __slots__ = ["uid", "loginType", "deleted", "iterations", "salt", "encryptedClientKey", "revision", "name"]
    UID_FIELD_NUMBER: _ClassVar[int]
    LOGINTYPE_FIELD_NUMBER: _ClassVar[int]
    DELETED_FIELD_NUMBER: _ClassVar[int]
    ITERATIONS_FIELD_NUMBER: _ClassVar[int]
    SALT_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDCLIENTKEY_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    uid: bytes
    loginType: _APIRequest_pb2.LoginType
    deleted: bool
    iterations: int
    salt: bytes
    encryptedClientKey: bytes
    revision: int
    name: str
    def __init__(self, uid: _Optional[bytes] = ..., loginType: _Optional[_Union[_APIRequest_pb2.LoginType, str]] = ..., deleted: bool = ..., iterations: _Optional[int] = ..., salt: _Optional[bytes] = ..., encryptedClientKey: _Optional[bytes] = ..., revision: _Optional[int] = ..., name: _Optional[str] = ...) -> None: ...

class BreachWatchSecurityData(_message.Message):
    __slots__ = ["recordUid", "revision"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    revision: int
    def __init__(self, recordUid: _Optional[bytes] = ..., revision: _Optional[int] = ...) -> None: ...

class ReusedPasswords(_message.Message):
    __slots__ = ["count", "revision"]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    count: int
    revision: int
    def __init__(self, count: _Optional[int] = ..., revision: _Optional[int] = ...) -> None: ...

class SharedFolderRecord(_message.Message):
    __slots__ = ["sharedFolderUid", "recordUid", "recordKey", "canShare", "canEdit", "ownerAccountUid", "expiration", "owner", "expirationNotificationType", "ownerUsername", "rotateOnExpiration"]
    SHAREDFOLDERUID_FIELD_NUMBER: _ClassVar[int]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    RECORDKEY_FIELD_NUMBER: _ClassVar[int]
    CANSHARE_FIELD_NUMBER: _ClassVar[int]
    CANEDIT_FIELD_NUMBER: _ClassVar[int]
    OWNERACCOUNTUID_FIELD_NUMBER: _ClassVar[int]
    EXPIRATION_FIELD_NUMBER: _ClassVar[int]
    OWNER_FIELD_NUMBER: _ClassVar[int]
    EXPIRATIONNOTIFICATIONTYPE_FIELD_NUMBER: _ClassVar[int]
    OWNERUSERNAME_FIELD_NUMBER: _ClassVar[int]
    ROTATEONEXPIRATION_FIELD_NUMBER: _ClassVar[int]
    sharedFolderUid: bytes
    recordUid: bytes
    recordKey: bytes
    canShare: bool
    canEdit: bool
    ownerAccountUid: bytes
    expiration: int
    owner: bool
    expirationNotificationType: _record_pb2.TimerNotificationType
    ownerUsername: str
    rotateOnExpiration: bool
    def __init__(self, sharedFolderUid: _Optional[bytes] = ..., recordUid: _Optional[bytes] = ..., recordKey: _Optional[bytes] = ..., canShare: bool = ..., canEdit: bool = ..., ownerAccountUid: _Optional[bytes] = ..., expiration: _Optional[int] = ..., owner: bool = ..., expirationNotificationType: _Optional[_Union[_record_pb2.TimerNotificationType, str]] = ..., ownerUsername: _Optional[str] = ..., rotateOnExpiration: bool = ...) -> None: ...

class SharedFolderUser(_message.Message):
    __slots__ = ["sharedFolderUid", "username", "manageRecords", "manageUsers", "accountUid", "expiration", "expirationNotificationType", "rotateOnExpiration"]
    SHAREDFOLDERUID_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    MANAGERECORDS_FIELD_NUMBER: _ClassVar[int]
    MANAGEUSERS_FIELD_NUMBER: _ClassVar[int]
    ACCOUNTUID_FIELD_NUMBER: _ClassVar[int]
    EXPIRATION_FIELD_NUMBER: _ClassVar[int]
    EXPIRATIONNOTIFICATIONTYPE_FIELD_NUMBER: _ClassVar[int]
    ROTATEONEXPIRATION_FIELD_NUMBER: _ClassVar[int]
    sharedFolderUid: bytes
    username: str
    manageRecords: bool
    manageUsers: bool
    accountUid: bytes
    expiration: int
    expirationNotificationType: _record_pb2.TimerNotificationType
    rotateOnExpiration: bool
    def __init__(self, sharedFolderUid: _Optional[bytes] = ..., username: _Optional[str] = ..., manageRecords: bool = ..., manageUsers: bool = ..., accountUid: _Optional[bytes] = ..., expiration: _Optional[int] = ..., expirationNotificationType: _Optional[_Union[_record_pb2.TimerNotificationType, str]] = ..., rotateOnExpiration: bool = ...) -> None: ...

class SharedFolderTeam(_message.Message):
    __slots__ = ["sharedFolderUid", "teamUid", "name", "manageRecords", "manageUsers", "expiration", "expirationNotificationType", "rotateOnExpiration"]
    SHAREDFOLDERUID_FIELD_NUMBER: _ClassVar[int]
    TEAMUID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    MANAGERECORDS_FIELD_NUMBER: _ClassVar[int]
    MANAGEUSERS_FIELD_NUMBER: _ClassVar[int]
    EXPIRATION_FIELD_NUMBER: _ClassVar[int]
    EXPIRATIONNOTIFICATIONTYPE_FIELD_NUMBER: _ClassVar[int]
    ROTATEONEXPIRATION_FIELD_NUMBER: _ClassVar[int]
    sharedFolderUid: bytes
    teamUid: bytes
    name: str
    manageRecords: bool
    manageUsers: bool
    expiration: int
    expirationNotificationType: _record_pb2.TimerNotificationType
    rotateOnExpiration: bool
    def __init__(self, sharedFolderUid: _Optional[bytes] = ..., teamUid: _Optional[bytes] = ..., name: _Optional[str] = ..., manageRecords: bool = ..., manageUsers: bool = ..., expiration: _Optional[int] = ..., expirationNotificationType: _Optional[_Union[_record_pb2.TimerNotificationType, str]] = ..., rotateOnExpiration: bool = ...) -> None: ...

class KsmChange(_message.Message):
    __slots__ = ["appRecordUid", "detailId", "removed", "appClientType", "expiration"]
    APPRECORDUID_FIELD_NUMBER: _ClassVar[int]
    DETAILID_FIELD_NUMBER: _ClassVar[int]
    REMOVED_FIELD_NUMBER: _ClassVar[int]
    APPCLIENTTYPE_FIELD_NUMBER: _ClassVar[int]
    EXPIRATION_FIELD_NUMBER: _ClassVar[int]
    appRecordUid: bytes
    detailId: bytes
    removed: bool
    appClientType: _enterprise_pb2.AppClientType
    expiration: int
    def __init__(self, appRecordUid: _Optional[bytes] = ..., detailId: _Optional[bytes] = ..., removed: bool = ..., appClientType: _Optional[_Union[_enterprise_pb2.AppClientType, str]] = ..., expiration: _Optional[int] = ...) -> None: ...

class ShareInvitation(_message.Message):
    __slots__ = ["username"]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    username: str
    def __init__(self, username: _Optional[str] = ...) -> None: ...

class User(_message.Message):
    __slots__ = ["accountUid", "username"]
    ACCOUNTUID_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    accountUid: bytes
    username: str
    def __init__(self, accountUid: _Optional[bytes] = ..., username: _Optional[str] = ...) -> None: ...

class SyncDiagnostics(_message.Message):
    __slots__ = ["continuationToken", "userId", "enterpriseUserId", "syncedTo", "syncingTo"]
    CONTINUATIONTOKEN_FIELD_NUMBER: _ClassVar[int]
    USERID_FIELD_NUMBER: _ClassVar[int]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    SYNCEDTO_FIELD_NUMBER: _ClassVar[int]
    SYNCINGTO_FIELD_NUMBER: _ClassVar[int]
    continuationToken: bytes
    userId: int
    enterpriseUserId: int
    syncedTo: int
    syncingTo: int
    def __init__(self, continuationToken: _Optional[bytes] = ..., userId: _Optional[int] = ..., enterpriseUserId: _Optional[int] = ..., syncedTo: _Optional[int] = ..., syncingTo: _Optional[int] = ...) -> None: ...

class RecordRotation(_message.Message):
    __slots__ = ["recordUid", "revision", "configurationUid", "schedule", "pwdComplexity", "disabled", "resourceUid", "lastRotation", "lastRotationStatus"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    CONFIGURATIONUID_FIELD_NUMBER: _ClassVar[int]
    SCHEDULE_FIELD_NUMBER: _ClassVar[int]
    PWDCOMPLEXITY_FIELD_NUMBER: _ClassVar[int]
    DISABLED_FIELD_NUMBER: _ClassVar[int]
    RESOURCEUID_FIELD_NUMBER: _ClassVar[int]
    LASTROTATION_FIELD_NUMBER: _ClassVar[int]
    LASTROTATIONSTATUS_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    revision: int
    configurationUid: bytes
    schedule: str
    pwdComplexity: bytes
    disabled: bool
    resourceUid: bytes
    lastRotation: int
    lastRotationStatus: RecordRotationStatus
    def __init__(self, recordUid: _Optional[bytes] = ..., revision: _Optional[int] = ..., configurationUid: _Optional[bytes] = ..., schedule: _Optional[str] = ..., pwdComplexity: _Optional[bytes] = ..., disabled: bool = ..., resourceUid: _Optional[bytes] = ..., lastRotation: _Optional[int] = ..., lastRotationStatus: _Optional[_Union[RecordRotationStatus, str]] = ...) -> None: ...

class SecurityScoreData(_message.Message):
    __slots__ = ["recordUid", "data", "revision"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    data: bytes
    revision: int
    def __init__(self, recordUid: _Optional[bytes] = ..., data: _Optional[bytes] = ..., revision: _Optional[int] = ...) -> None: ...

class BreachWatchGetSyncDataRequest(_message.Message):
    __slots__ = ["recordUids"]
    RECORDUIDS_FIELD_NUMBER: _ClassVar[int]
    recordUids: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, recordUids: _Optional[_Iterable[bytes]] = ...) -> None: ...

class BreachWatchGetSyncDataResponse(_message.Message):
    __slots__ = ["breachWatchRecords", "breachWatchSecurityData", "users"]
    BREACHWATCHRECORDS_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHSECURITYDATA_FIELD_NUMBER: _ClassVar[int]
    USERS_FIELD_NUMBER: _ClassVar[int]
    breachWatchRecords: _containers.RepeatedCompositeFieldContainer[BreachWatchRecord]
    breachWatchSecurityData: _containers.RepeatedCompositeFieldContainer[BreachWatchSecurityData]
    users: _containers.RepeatedCompositeFieldContainer[User]
    def __init__(self, breachWatchRecords: _Optional[_Iterable[_Union[BreachWatchRecord, _Mapping]]] = ..., breachWatchSecurityData: _Optional[_Iterable[_Union[BreachWatchSecurityData, _Mapping]]] = ..., users: _Optional[_Iterable[_Union[User, _Mapping]]] = ...) -> None: ...

class GetAccountUidMapResponse(_message.Message):
    __slots__ = ["users"]
    USERS_FIELD_NUMBER: _ClassVar[int]
    users: _containers.RepeatedCompositeFieldContainer[User]
    def __init__(self, users: _Optional[_Iterable[_Union[User, _Mapping]]] = ...) -> None: ...



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import signal
import sys
from ..decorators.logging import logger

class SignalHandler:
    @staticmethod
    def setup_signal_handlers(shutdown_callback) -> None:
        """Set up signal handlers for graceful shutdown."""
        def signal_handler(signum, frame):
            logger.debug("Received shutdown signal")
            shutdown_callback()
            sys.exit(0)
            
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from flask import Blueprint, request, jsonify, Response
import queue
from typing import Tuple, Union
from ..decorators.unified import unified_api_decorator
from ..util.command_util import CommandExecutor
from ..decorators.logging import logger
from ..core.request_queue import queue_manager
from ..util.request_validation import RequestValidator

def create_legacy_command_blueprint():
    """Create legacy blueprint for direct/synchronous command execution (non-queue mode)."""
    bp = Blueprint("legacy_command_bp", __name__)
    
    @bp.after_request
    def add_legacy_header(response):
        """Add legacy header for legacy API."""
        response.headers['X-API-Legacy'] = 'true'
        return response
    
    @bp.route("/executecommand", methods=["POST"])
    @unified_api_decorator()
    def execute_command_direct(**kwargs) -> Tuple[Union[Response, bytes], int]:
        """Execute command directly and return result immediately (V1 API behavior)."""
        temp_files = []
        try:
            json_error = RequestValidator.validate_request_json()
            if json_error:
                return json_error
            
            command, validation_error = RequestValidator.validate_and_escape_command(request.json)
            if validation_error:
                return validation_error
            
            # Process file data if present
            processed_command, temp_files = RequestValidator.process_file_data(request.json, command)
                
            response, status_code = CommandExecutor.execute(processed_command)
            
            # If we get a busy response, add v1-specific message
            if (isinstance(response, dict) and 
                "temporarily busy" in str(response.get("error", "")).lower()):
                response["message"] = "Note: api/v1/executecommand only supports a single request at a time."
                status_code = 503
            
            return response if isinstance(response, bytes) else jsonify(response), status_code

        except Exception as e:
            logger.error(f"Error executing command: {e}")
            return jsonify({"status": "error", "error": f"Error: {str(e)}"}), 500
        finally:
            # Clean up temporary files
            RequestValidator.cleanup_temp_files(temp_files)

    return bp

def create_command_blueprint():
    """Create Blue Print for Keeper Commander Service."""
    bp = Blueprint("command_bp", __name__)
    
    @bp.route("/executecommand-async", methods=["POST"])
    @unified_api_decorator()
    def execute_command(**kwargs) -> Tuple[Response, int]:
        """Submit a command for execution and return request ID immediately."""
        temp_files = []
        try:
            json_error = RequestValidator.validate_request_json()
            if json_error:
                return json_error
            
            command, validation_error = RequestValidator.validate_and_escape_command(request.json)
            if validation_error:
                return validation_error
            
            # Process file data if present
            processed_command, temp_files = RequestValidator.process_file_data(request.json, command)
            
            # Submit to queue and return request ID immediately
            try:
                request_id = queue_manager.submit_request(processed_command, temp_files)
                return jsonify({
                    "success": True,
                    "request_id": request_id,
                    "status": "queued",
                    "message": "Request queued successfully. Use /api/v2/status/<request_id> to check progress, /api/v2/result/<request_id> to get results, or /api/v2/queue/status for queue info."
                }), 202  # 202 Accepted
            except queue.Full:
                # Clean up temp files if queue is full
                RequestValidator.cleanup_temp_files(temp_files)
                return jsonify({
                    "status": "error", 
                    "error": "Error: Request queue is full. Please try again later."
                }), 503  # 503 Service Unavailable
                
        except Exception as e:
            logger.error(f"Error submitting request: {e}")
            # Clean up temp files on error
            RequestValidator.cleanup_temp_files(temp_files)
            return jsonify({"status": "error", "error": f"{str(e)}"}), 500

    @bp.route("/status/<request_id>", methods=["GET"])
    @unified_api_decorator()
    def get_request_status(request_id: str, **kwargs) -> Tuple[Response, int]:
        """Get the status of a specific request."""
        try:
            status_info = queue_manager.get_request_status(request_id)
            if status_info is None:
                return jsonify({
                    "status": "error",
                    "error": "Error: Request not found"
                }), 404
                
            return jsonify({
                "success": True,
                "request_id": request_id,
                **status_info
            }), 200
            
        except Exception as e:
            logger.error(f"Error getting request status: {e}")
            return jsonify({"status": "error", "error": str(e)}), 500

    @bp.route("/result/<request_id>", methods=["GET"])
    @unified_api_decorator()
    def get_request_result(request_id: str, **kwargs) -> Tuple[Union[Response, bytes], int]:
        """Get the result of a completed request."""
        try:
            result_data = queue_manager.get_request_result(request_id)
            if result_data is None:
                # Check if request exists but isn't completed
                status_info = queue_manager.get_request_status(request_id)
                if status_info is None:
                    return jsonify({
                        "status": "error",
                        "error": "Error: Request not found"
                    }), 404
                else:
                    return jsonify({
                        "status": "error",
                        "error": "Error: Request not completed yet",
                        "status": status_info["status"]
                    }), 202  # 202 Accepted (still processing)
            
            result, status_code = result_data
            return result if isinstance(result, bytes) else jsonify(result), status_code
            
        except Exception as e:
            logger.error(f"Error getting request result: {e}")
            return jsonify({"status": "error", "error": str(e)}), 500

    @bp.route("/queue/status", methods=["GET"])
    @unified_api_decorator()
    def get_queue_status(**kwargs) -> Tuple[Response, int]:
        """Get overall queue status information."""
        try:
            queue_status = queue_manager.get_queue_status()
            return jsonify({
                "success": True,
                **queue_status
            }), 200
            
        except Exception as e:
            logger.error(f"Error getting queue status: {e}")
            return jsonify({"status": "error", "error": str(e)}), 500

    return bp



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import json
import logging
import base64
import os
import datetime
from typing import Optional, List, Dict, Any

from .base import GroupCommand, dump_report_data, report_output_parser, field_to_title, user_choice
from .enterprise_common import EnterpriseCommand
from .. import api, utils
from ..display import bcolors
from ..error import CommandError
from ..params import KeeperParams
from ..proto import publicapi_pb2

api_key_list_parser = argparse.ArgumentParser(
    prog='public-api-key list', 
    parents=[report_output_parser],
    description='Display a list of enterprise API keys.',
    epilog='''
Examples:
  # List all API keys in table format
  public-api-key list
  
  # List API keys in JSON format
  public-api-key list --format json
  
  # Save API keys list to CSV file
  public-api-key list --format csv --output api_keys.csv
    ''',
    formatter_class=argparse.RawDescriptionHelpFormatter
)

api_key_generate_parser = argparse.ArgumentParser(
    prog='public-api-key generate', 
    description='Generate a new enterprise API key.',
    epilog='''
Examples:
  # Generate API key with SIEM role and 30-day expiration
  public-api-key generate --name "SIEM Integration" --roles "SIEM:2" --expires 30d
  
  # Generate API key with multiple roles and 24-hour expiration
  public-api-key generate --name "Temp Access" --roles "Admin:1,SIEM:2" --expires 24h
  
  # Generate permanent API key with read-only access
  public-api-key generate --name "Monitoring Tool" --roles "ReadOnly:1" --expires never
  
  # Generate API key and save details to JSON file
  public-api-key generate --name "Backup Tool" --roles "Backup:2" --expires 1y --format json --output backup_key.json
    ''',
    formatter_class=argparse.RawDescriptionHelpFormatter
)
api_key_generate_parser.add_argument('--name', dest='name', required=True, 
                                     help='API key name. Examples: "SIEM Integration", "Backup Tool", "Monitoring Service"')
api_key_generate_parser.add_argument('--roles', dest='roles', required=True, action='store', 
                                     help='''Comma-separated list of role IDs with action types. 
Format: "RoleID:ActionType" or "RoleName:ActionType"
Action types: 0=NONE (no permissions), 1=READ (read-only), 2=READ_WRITE (full access)
Examples: 
  --roles "SIEM:2"                    # SIEM role with read-write access
  --roles "Admin:1,SIEM:2"            # Multiple roles with different permissions
  --roles "123:1,456:2"               # Using role IDs instead of names
  --roles "ReadOnly:1"                # Read-only access
  --roles "Backup:2,Monitor:1"        # Backup with full access, Monitor with read-only''')
api_key_generate_parser.add_argument('--expires', dest='expires', action='store',
                                     choices=['24h', '7d', '30d', '1y', 'never'],
                                     default='never',
                                     help='''Expiration time for the API key:
  24h   = 24 hours from now (temporary access)
  7d    = 7 days from now (short-term projects)
  30d   = 30 days from now (monthly integrations)
  1y    = 1 year from now (long-term integrations)
  never = never expires (permanent access, use with caution)
Default: never''')
api_key_generate_parser.add_argument('--format', dest='format', action='store', choices=['table', 'json'], default='table',
                                     help='''Output format:
  table = human-readable table format (default)
  json  = JSON format for programmatic use''')
api_key_generate_parser.add_argument('--output', dest='output', action='store',
                                     help='''Output file name (only used with --format json).
Examples: 
  --output api_key.json
  --output /path/to/keys/siem_key.json
If not specified, JSON output is printed to console.''')

api_key_revoke_parser = argparse.ArgumentParser(
    prog='public-api-key revoke', 
    description='Revoke an enterprise API key.',
    epilog='''
Examples:
  # Revoke API key with confirmation prompt
  public-api-key revoke 12345
  
  # Revoke API key without confirmation (force)
  public-api-key revoke 12345 --force
  
  # Short form with force flag
  public-api-key revoke 12345 -f

Tips:
  - Use 'public-api-key list' to find the Token you want to revoke
  - Revoked keys cannot be restored, only new keys can be generated
  - Use --force to skip the confirmation prompt (useful for scripts)
    ''',
    formatter_class=argparse.RawDescriptionHelpFormatter
)
api_key_revoke_parser.add_argument('token', 
                                   help='API key Token (get this from "api-key list" command). Example: 12345')
api_key_revoke_parser.add_argument('--force', '-f', dest='force', action='store_true', 
                                   help='Revoke without confirmation prompt (useful for automated scripts)')


def register_commands(commands):
    commands['public-api-key'] = ApiKeyCommand()


def register_command_info(aliases, command_info):
    command_info['public-api-key'] = 'Manage Admin REST API keys for 3rd party integrations'


class ApiKeyCommand(GroupCommand):
    def __init__(self):
        super(ApiKeyCommand, self).__init__()
        self.register_command('list', ApiKeyListCommand())
        self.register_command('generate', ApiKeyGenerateCommand())
        self.register_command('revoke', ApiKeyRevokeCommand())
        self.default_verb = 'list'
    
    def print_help(self, **kwargs):
        print('Enterprise API Key Management')
        print('=' * 50)
        print()
        print('Commands:')
        print('  list      - Display all enterprise API keys')
        print('  generate  - Create a new API key with specified roles and expiration')
        print('  revoke    - Revoke an existing API key')
        print()
        print('Quick Start Examples:')
        print('  # List all API keys')
        print('  public-api-key list')
        print()
        print('  # Generate a new API key for SIEM integration (30-day expiration)')
        print('  public-api-key generate --name "SIEM Tool" --roles "SIEM:2" --expires 30d')
        print()
        print('  # Revoke an API key')
        print('  public-api-key revoke 12345')
        print()
        print('Role Action Types:')
        print('  1 = READ       (read-only access)')
        print('  2 = READ_WRITE (full access)')
        print()
        print('Expiration Options:')
        print('  24h   = 24 hours    7d = 7 days    30d = 30 days')
        print('  1y    = 1 year      never = permanent (use with caution)')
        print()
        print('For detailed help on any command, use:')
        print('  public-api-key <command> --help')
        print('  Example: public-api-key generate --help')
        print()


class ApiKeyListCommand(EnterpriseCommand):
    def get_parser(self):
        return api_key_list_parser

    def execute(self, params, **kwargs):
        fmt = kwargs.get('format') or ''
        headers = ['token', 'enterprise_id', 'name', 'status', 'issued_date', 'expiration_date', 'roles']
        if fmt != 'json':
            headers = [field_to_title(x) for x in headers]

        table = []
        try:
            # Create request for listing tokens
            rq = publicapi_pb2.ListPublicApiTokenRequest()
            rq.statusFilter = publicapi_pb2.StatusFilter.ALL
            
            # Call the REST API
            rs = api.communicate_rest(
                params, rq, 
                'public_api/list_token',
                rs_type=publicapi_pb2.PublicApiTokenResponseList
            )
            
            for token in rs.tokens:
                issued_date = ''
                if token.issuedDate:
                    dt = datetime.datetime.fromtimestamp(token.issuedDate / 1000)
                    issued_date = dt.strftime('%Y-%m-%d %H:%M:%S')
                
                expiration_date = 'Never'
                if token.expirationDate:
                    dt = datetime.datetime.fromtimestamp(token.expirationDate / 1000)
                    expiration_date = dt.strftime('%Y-%m-%d %H:%M:%S')
                
                roles_str = ', '.join([f"{integration.roleName}:{integration.actionType}" 
                                     for integration in token.integrations])
                
                if token.expirationDate and token.expirationDate < int(datetime.datetime.now().timestamp() * 1000):
                    status = 'Expired'
                else:
                    status = 'Active'
                
                
                row = [
                    token.token,
                    token.enterprise_id,
                    token.name,
                    status,
                    issued_date,
                    expiration_date,
                    roles_str
                ]
                table.append(row)
                
        except Exception as e:
            logging.error(f"Failed to list API keys: {e}")
            raise CommandError("Failed to retrieve API keys")
        
        return dump_report_data(table, headers=headers, fmt=fmt, filename=kwargs.get('output'))


class ApiKeyGenerateCommand(EnterpriseCommand):
    def get_parser(self):
        return api_key_generate_parser

    def execute(self, params, **kwargs):
        name = kwargs.get('name')
        if not name:
            print("API key name is required")
            return

        try:
            # Create the generate token request
            rq = publicapi_pb2.GenerateTokenRequest()
            rq.tokenName = name
            rq.issuedDate = int(datetime.datetime.now().timestamp() * 1000)
            
            # Set expiration based on the selected option
            expires = kwargs.get('expires', 'never')
            if expires != 'never':
                now = datetime.datetime.now()
                if expires == '24h':
                    expiration_date = now + datetime.timedelta(hours=24)
                elif expires == '7d':
                    expiration_date = now + datetime.timedelta(days=7)
                elif expires == '30d':
                    expiration_date = now + datetime.timedelta(days=30)
                elif expires == '1y':
                    expiration_date = now + datetime.timedelta(days=365)
                else:
                    print(f"Invalid expiration option: {expires}")
                    return
                
                rq.expirationDate = int(expiration_date.timestamp() * 1000)
            # If expires == 'never', don't set expirationDate (it remains unset/blank)
            
            # Parse roles - now required
            roles_str = kwargs.get('roles')
            if not roles_str:
                print("At least one role is required. Example: --roles 'SIEM:2,CSPM:1'")
                return
            
            for role_spec in roles_str.split(','):
                role_spec = role_spec.strip()
                
                # Require format: "RoleName:ActionType" or "RoleID:ActionType"

                if ':' in role_spec:
                    role_id_str, action_type_str = role_spec.split(':', 1)
                    allowed_roles = [("SIEM", 1), ("CSPM", 2), ("BILLING", 3)]
                    allowed_role_names = [role[0].upper() for role in allowed_roles]
                    if role_id_str.strip().upper() not in allowed_role_names:
                        print(f"Role '{role_id_str.strip()}' does not match allowed roles: {', '.join(allowed_role_names)}. Skipping.")
                        return
                    role_id_str = role_id_str.strip()
                    role_id = next(role[1] for role in allowed_roles if role[0].upper() == role_id_str.upper())
                    action_type_str = action_type_str.strip()
                else:
                    # If no action type specified, default to READ-write (2)
                    print(f"Error: Role specification must include action type. Got: '{role_spec}'")
                    print("Required format: 'RoleName:ActionType' (e.g., 'SIEM:1,CSPM:2,Billing:1')")
                    return
                
                # Map action type number to enum
                try:
                    action_type_num = int(action_type_str)
                    if action_type_num == 1:
                        action_type = publicapi_pb2.ActionType.READ
                    elif action_type_num == 2:
                        action_type = publicapi_pb2.ActionType.READ_WRITE
                    else:
                        print(f"Invalid action type: '{action_type_str}'. Valid values are: 1=READ, 2=READ_WRITE. Defaulting to READ-write (2)")
                        action_type = publicapi_pb2.ActionType.READ_WRITE
                except ValueError:
                    print(f"Invalid action type: '{action_type_str}'. Action type must be a number: 1=READ, 2=READ_WRITE. Defaulting to read-write (2)")
                    action_type = publicapi_pb2.ActionType.READ_WRITE
                
                role = publicapi_pb2.Role()
                role.roleId = role_id
                role.actionType = action_type
                rq.roles.append(role)
            
            # Send the request
            rs = api.communicate_rest(
                params, rq,
                'public_api/generate_token',
                rs_type=publicapi_pb2.PublicApiTokenResponse
            )
            
            fmt = kwargs.get('format') or 'table'
            if fmt == 'json':
                output = {
                    'name': rs.name,
                    'token': rs.token,
                    'enterprise_id': rs.enterprise_id,
                    'issued_date': rs.issuedDate,
                    'expiration_date': rs.expirationDate if rs.expirationDate else 'never',
                    'integrations': [
                        {
                            'role_name': integration.roleName,
                            'action_type': integration.actionType,
                            'action_type_name': publicapi_pb2.ActionType.Name(integration.actionType)
                        } for integration in rs.integrations
                    ]
                }
                
                output_file = kwargs.get('output')
                if output_file:
                    with open(output_file, 'w') as f:
                        json.dump(output, f, indent=2)
                    print('File name: %s', os.path.abspath(output_file))
                else:
                    return json.dumps(output, indent=2)
            else:
                print(f"{bcolors.OKGREEN}API Key generated successfully{bcolors.ENDC}")
                print(f"Token: {rs.token}")
                print(f"Name: {rs.name}")
                print(f"Token: {rs.token}")
                print(f"Enterprise ID: {rs.enterprise_id}")
                if rs.expirationDate:
                    exp_date = datetime.datetime.fromtimestamp(rs.expirationDate / 1000)
                    print(f"Expires: {exp_date.strftime('%Y-%m-%d %H:%M:%S')}")
                else:
                    print("Expires: Never")
                
                if rs.integrations:
                    print("Roles:")
                    for integration in rs.integrations:
                        action_name = publicapi_pb2.ActionType.Name(integration.actionType)
                        print(f"  - {integration.roleName}: {action_name} ({integration.actionType})")
                        
        except Exception as e:
            logging.error(f"Failed to generate API key: {e}")
            raise CommandError("Failed to generate API key")


class ApiKeyRevokeCommand(EnterpriseCommand):
    def get_parser(self):
        return api_key_revoke_parser

    def execute(self, params, **kwargs):
        token = kwargs.get('token')
        if not token:
            print("Token is required")
            return
        
        # Confirm revocation unless force flag is set
        if not kwargs.get('force'):
            answer = user_choice(
                bcolors.FAIL + bcolors.BOLD + '\nALERT!\n' + bcolors.ENDC +
                f'You are about to revoke API key with Token {token}' +
                '\n\nDo you want to proceed with revocation?', 'yn', 'n'
            )
            if answer.lower() != 'y':
                return
        
        try:
            # Create the revoke token request
            rq = publicapi_pb2.RevokeTokenRequest()
            rq.token = token
            
            # Send the request
            rs = api.communicate_rest(
                params, rq,
                'public_api/revoke_token',
                rs_type=publicapi_pb2.RevokeTokenResponse
            )
            
            print(f"{bcolors.OKGREEN}API Key with Token {token} revoked successfully{bcolors.ENDC}")
            if rs.message:
                print(f"Message: {rs.message}")
                
        except Exception as e:
            logging.error(f"Failed to revoke API key: {e}")
            raise CommandError("Failed to revoke API key") 



import sys
if sys.version_info >= (3, 8):
    import pytest
    from unittest import TestCase, mock
    from flask import Flask
    from keepercommander.service.decorators.auth import auth_check, policy_check
    from keepercommander.service.decorators.security import security_check, is_allowed_ip
    from keepercommander.service.util.config_reader import ConfigReader

    class TestAuthSecurity(TestCase):
        def setUp(self):
            self.app = Flask(__name__)
            self.client = self.app.test_client()
            
            @self.app.route('/test', methods=['POST'])
            @security_check
            @auth_check
            @policy_check
            def test_endpoint():
                return {'status': 'success'}, 200

        def test_auth_check_missing_api_key(self):
            """Test authentication with missing API key"""
            with self.app.test_request_context('/test', method='POST'):
                response = auth_check(lambda *args, **kwargs: ({'status': 'success'}, 200))()
                self.assertEqual(response[1], 401)
                self.assertEqual(response[0]['status'], 'error')
                self.assertIn('api key', response[0]['error'])

        @mock.patch.object(ConfigReader, 'read_config')
        def test_auth_check_invalid_api_key(self, mock_read_config):
            """Test authentication with invalid API key"""
            mock_read_config.return_value = "different_key"
            
            with self.app.test_request_context('/test', method='POST', 
                                            headers={'api-key': 'test_key'}):
                response = auth_check(lambda *args, **kwargs: ({'status': 'success'}, 200))()
                self.assertEqual(response[1], 401)
                self.assertEqual(response[0]['status'], 'error')

        @mock.patch.object(ConfigReader, 'read_config')
        def test_auth_check_expired_key(self, mock_read_config):
            """Test authentication with expired API key"""
            mock_read_config.side_effect = [
                "test_key",
                "2024-01-01T00:00:00"
            ]
            
            with self.app.test_request_context('/test', method='POST', 
                                            headers={'api-key': 'test_key'}):
                response = auth_check(lambda *args, **kwargs: ({'status': 'success'}, 200))()
                self.assertEqual(response[1], 401)
                self.assertEqual(response[0]['status'], 'error')
                self.assertIn('expired', response[0]['error'])

        # def test_security_check_blocked_ip(self):
        #     """Test security check with blocked IP"""
        #     with mock.patch.object(ConfigReader, 'read_config', return_value="192.168.1.1"):
        #         with self.app.test_request_context('/test', method='POST', 
        #                                         environ_base={'REMOTE_ADDR': '192.168.1.1'}):
        #             response = security_check(lambda *args, **kwargs: ({'status': 'success'}, 200))()
        #             response_data = response[0].get_json()
        #             self.assertEqual(response[1], 403) 
        #             self.assertEqual(response_data['error'], 'IP is blocked')

        def test_is_blocked_ip_single_ip(self):
            """Test IP blocking with single IP address"""
            blocked_ips = "192.168.1.1"
            allowed_ips="192.168.1.2"
            self.assertFalse(is_allowed_ip("192.168.1.1", allowed_ips, blocked_ips))
            self.assertTrue(is_allowed_ip("192.168.1.2", allowed_ips, blocked_ips))

        def test_is_blocked_ip_cidr(self):
            """Test IP blocking with CIDR notation"""
            allowed_ips="192.168.1.1"
            blocked_ips = "192.168.1.0"
            self.assertTrue(is_allowed_ip("192.168.1.1", allowed_ips, blocked_ips))
            self.assertFalse(is_allowed_ip("192.168.1.254", allowed_ips, blocked_ips))
            self.assertFalse(is_allowed_ip("192.168.2.1", allowed_ips, blocked_ips))

        @mock.patch.object(ConfigReader, 'read_config')
        def test_policy_check_allowed_command(self, mock_read_config):
            """Test policy check with allowed command"""
            mock_read_config.return_value = "list,get,search"
            
            with self.app.test_request_context('/test', method='POST',
                                            json={"command": "list"}):
                response = policy_check(lambda *args, **kwargs: ({'status': 'success'}, 200))()
                self.assertEqual(response[1], 200)
                self.assertEqual(response[0]['status'], 'success')

        @mock.patch.object(ConfigReader, 'read_config')
        def test_policy_check_denied_command(self, mock_read_config):
            """Test policy check with denied command"""
            mock_read_config.return_value = "list,get,search"
            
            with self.app.test_request_context('/test', method='POST',
                                            json={"command": "delete"}):
                response = policy_check(lambda *args, **kwargs: ({'status': 'success'}, 200))()
                self.assertEqual(response[1], 403)
                self.assertEqual(response[0]['status'], 'error')
                self.assertIn('Not permitted', response[0]['error'])


from typing import Sequence, Optional, List

from keeper_secrets_manager_core.utils import url_safe_str_to_bytes

from ... import api, utils
from ...commands.utils import KSMCommand
from ...loginv3 import CommonHelperMethods
from ...params import KeeperParams
from ...proto import pam_pb2, enterprise_pb2


def find_one_gateway_by_uid_or_name(params, gateway_name_or_uid):
    all_gateways = get_all_gateways(params)
    gateway_uid_bytes = url_safe_str_to_bytes(gateway_name_or_uid)

    found_gateways = list(filter(lambda g: g.controllerUid == gateway_uid_bytes or g.controllerName == gateway_name_or_uid, all_gateways))
    if not found_gateways:
        return None

    first_found_gateway_uid_bytes = found_gateways[0].controllerUid

    # TODO: Print warning if there are more than 1 gateway found
    found_gateway_uid_str = CommonHelperMethods.bytes_to_url_safe_str(first_found_gateway_uid_bytes)

    return found_gateway_uid_str


def get_all_gateways(params):  # type: (KeeperParams) -> Sequence[pam_pb2.PAMController]
    rs = api.communicate_rest(params, None, 'pam/get_controllers', rs_type=pam_pb2.PAMControllersResponse)
    return rs.controllers


def find_connected_gateways(all_controllers, identifier):  # type: (List[bytes], str) -> Optional[bytes]
    # Will search connected controllers by the "controllerName", "deviceName", or "deviceToken"

    found_connected_controller_uid_bytes = next((c for c in all_controllers if (utils.base64_url_encode(c) == identifier)), None)

    if found_connected_controller_uid_bytes:
        # if len(found_connected_controller) > 1:
        #     logging.warning(f"More than one gateway with the same identifier [{identifier}] was located.")
        # else:
        #     return found_connected_controller[0]
        return found_connected_controller_uid_bytes
    else:
        return None


def create_gateway(params, gateway_name, ksm_app, config_init, ott_expire_in_min=5):

    one_time_token_dict = KSMCommand.add_client(params,
                                                app_name_or_uid=ksm_app,
                                                count=1,
                                                unlock_ip=True,
                                                first_access_expire_on=ott_expire_in_min,  # if one time token not used in 5 min then it will be expired
                                                access_expire_in_min=None, # how long the client has access to the application, None=Never, int = num of min
                                                client_name=gateway_name,
                                                config_init=False,
                                                silent=True,
                                                client_type=enterprise_pb2.DISCOVERY_AND_ROTATION_CONTROLLER)

    one_time_token_dict = one_time_token_dict[0]
    one_time_token = one_time_token_dict.get('oneTimeToken')

    if config_init:
        config_str_and_config_dict = KSMCommand.init_ksm_config(params, one_time_token, config_init,
                                                                include_config_dict=True)

        one_time_token = config_str_and_config_dict.get('config_str')

    return one_time_token


def remove_gateway(params, gateway_uid):   # type: (KeeperParams, bytes) -> None
    rq = pam_pb2.PAMGenericUidRequest()
    rq.uid = gateway_uid
    rs = api.communicate_rest(params, rq, 'pam/remove_controller', rs_type=pam_pb2.PAMRemoveControllerResponse)
    controller = next((x for x in rs.controllers if x.controllerUid == gateway_uid), None)
    if controller:
        raise Exception(controller.message)



#!/usr/bin/env python3
"""
Docker KSM Utility - Centralized utility for KSM operations in Docker containers.

This module provides secure file operations for KSM (Keeper Secrets Manager)
records, including config processing, file upload/download, and monitoring.
"""

import sys
import os
import argparse
import time
import hashlib
from pathlib import Path
import base64
import binascii
import glob
import json
import shutil
import tempfile

# =============================================================================
# SECURITY AND VALIDATION FUNCTIONS
# =============================================================================

def validate_file_path(file_path, base_dir=None):
    """
    Validate file path to prevent directory traversal attacks.
    
    Args:
        file_path (str): Path to validate
        base_dir (str, optional): Base directory to restrict access to
        
    Returns:
        tuple: (is_valid, resolved_path)
    """
    try:
        # Convert to Path object and resolve
        path = Path(file_path).resolve()
        
        # Check for directory traversal attempts
        if '..' in str(path) or str(path).startswith('/..'):
            return False, None
            
        # If base_dir is specified, ensure path is within it
        if base_dir:
            base_path = Path(base_dir).resolve()
            try:
                path.relative_to(base_path)
            except ValueError:
                return False, None
        
        # Additional security checks
        str_path = str(path)
        dangerous_patterns = [
            '../', '..\\', '~/', '/etc/', '/proc/', '/sys/'
        ]
        if any(pattern in str_path for pattern in dangerous_patterns):
            return False, None
            
        return True, str(path)
        
    except (OSError, ValueError):
        return False, None

def check_ksm_dependency():
    """Check if keeper_secrets_manager_core is installed.
    
    Returns:
        bool: True if installed, False otherwise
    """
    try:
        import keeper_secrets_manager_core  # noqa: F401
        return True
    except ImportError:
        print("ERROR: keeper_secrets_manager_core is not installed")
        return False


# =============================================================================
# KSM CONFIG PROCESSING FUNCTIONS
# =============================================================================

def is_base64_config(input_str):
    """
    Detect if input is base64-encoded JSON or file path.

    Args:
        input_str (str): Input string to check

    Returns:
        bool: True if base64-encoded JSON, False otherwise
    """
    if not input_str:
        return False
    
    # If it looks like a file path (starts with path indicators or exists as file)
    if (input_str.startswith('/') or input_str.startswith('./') or 
        input_str.startswith('../') or input_str.startswith('~') or 
        os.path.isfile(input_str)):
        return False

    try:
        decoded_bytes = base64.b64decode(input_str, validate=True)
        decoded_str = decoded_bytes.decode('utf-8')
        json.loads(decoded_str)  # Validate JSON
        return True
    except (binascii.Error, UnicodeDecodeError, json.JSONDecodeError):
        return False

def cleanup_old_ksm_dirs():
    """
    Clean up old KSM config directories.
    Removes old timestamped directories to prevent accumulation.
    """
    patterns = ["/home/commander/ksm_*", os.path.expanduser("~/ksm_*")]
    
    for pattern in patterns:
        try:
            ksm_dirs = glob.glob(pattern)
            
            if not ksm_dirs:
                continue
            
            # Remove old timestamped directories
            for old_dir in ksm_dirs:
                try:
                    shutil.rmtree(old_dir)
                except Exception as e:
                    print(f"Warning: Could not remove directory {old_dir}: {e}")
        except Exception:
            pass  # Ignore pattern errors

def _create_temp_ksm_dir():
    """
    Create temporary directory for KSM config with proper fallback.
    
    Returns:
        str: Path to created directory, or None if failed
    """
    try:
        # Try Docker path first
        if os.path.exists("/home/commander"):
            return tempfile.mkdtemp(prefix="ksm_", dir="/home/commander")
        else:
            # Fallback for local testing
            return tempfile.mkdtemp(prefix="ksm_")
    except Exception as e:
        print(f"ERROR: Failed to create temp directory: {e}")
        return None


def _decode_and_save_config(base64_input, config_path):
    """
    Decode base64 config and save to file.
    
    Args:
        base64_input (str): Base64-encoded JSON config
        config_path (str): Path where to save the config
        
    Returns:
        bool: True if successful, False otherwise
    """
    try:
        # Decode base64 and validate JSON
        decoded_bytes = base64.b64decode(base64_input)
        decoded_str = decoded_bytes.decode('utf-8')
        config_data = json.loads(decoded_str)
        
        # Basic validation of config structure
        if not isinstance(config_data, dict):
            print("ERROR: Invalid config format - must be JSON object")
            return False
        
        # Write to file with proper formatting
        with open(config_path, 'w', encoding='utf-8') as f:
            json.dump(config_data, f, indent=2, ensure_ascii=False)
        
        # Set secure permissions
        os.chmod(config_path, 0o600)
        return True
        
    except Exception as e:
        print("ERROR: Failed to decode and save config")
        return False


def process_ksm_config(ksm_config_input):
    """
    Process KSM config input - detect if base64 or file path and handle accordingly.
    
    Args:
        ksm_config_input (str): Either file path or base64-encoded JSON
        
    Returns:
        str: Final file path to use for KSM config, or None if failed
    """
    if not ksm_config_input:
        return None
    
    # Check if it's base64 or file path
    if is_base64_config(ksm_config_input):
        # Clean up any old KSM directories
        cleanup_old_ksm_dirs()
        
        # Create temporary directory
        ksm_dir = _create_temp_ksm_dir()
        if not ksm_dir:
            return None
            
        config_path = os.path.join(ksm_dir, "ksm-config.json")
        
        # Decode and save config
        if _decode_and_save_config(ksm_config_input, config_path):
            return config_path
        else:
            # Clean up on failure
            try:
                shutil.rmtree(ksm_dir)
            except Exception:
                pass
            return None
    else:
        # It's a file path, validate it exists
        
        if os.path.isfile(ksm_config_input):
            return ksm_config_input
        else:
            print("ERROR: KSM config file not found")
            return None


# =============================================================================
# KSM SECRETS MANAGER OPERATIONS
# =============================================================================

def _initialize_secrets_manager(ksm_config_path, ksm_token):
    """
    Initialize SecretsManager with proper error handling.
    
    Args:
        ksm_config_path (str): Path to KSM config file (optional)
        ksm_token (str): KSM access token (optional)
        
    Returns:
        SecretsManager: Initialized SecretsManager or None if failed
    """
    from keeper_secrets_manager_core import SecretsManager
    from keeper_secrets_manager_core.storage import FileKeyValueStorage
    
    try:
        if ksm_config_path:
            if not os.path.exists(ksm_config_path):
                print("ERROR: KSM config file not found")
                return None
            return SecretsManager(config=FileKeyValueStorage(ksm_config_path))
        else:
            return SecretsManager(token=ksm_token)
    except Exception as e:
        print(f"ERROR: Failed to initialize SecretsManager: {e}")
        return None


def _get_secret_by_uid_or_title(secrets_manager, record_identifier):
    """
    Get secret by UID first, then by title as fallback.
    
    Args:
        secrets_manager: Initialized SecretsManager instance
        record_identifier (str): Record UID or title
        
    Returns:
        secret_object: Secret object or None if not found
    """
    try:
        # First try to get by UID
        secrets = secrets_manager.get_secrets([record_identifier])
        if secrets and len(secrets) > 0:
            return secrets[0]
    except Exception:
        # UID lookup failed, continue to title lookup
        pass
    
    try:
        # Try to get by title as fallback
        secrets = secrets_manager.get_secrets_by_title(record_identifier)
        
        if not secrets or len(secrets) == 0:
            print(f"ERROR: Record not found by UID or title: {record_identifier}")
            return None
        elif len(secrets) > 1:
            print(f"ERROR: Multiple records found with title '{record_identifier}' "
                  f"({len(secrets)} records). Please use UID or a unique title.")
            return None
        else:
            return secrets[0]
            
    except Exception as e:
        print(f"ERROR: Failed to lookup record by title: {e}")
        pass
    
    print(f"ERROR: Record not found by UID or title: {record_identifier}")
    return None


def download_config(ksm_config_path, ksm_token, record_identifier, output_path):
    """
    Download config.json from KSM record.
    
    Args:
        ksm_config_path (str): Path to KSM config file (optional)
        ksm_token (str): KSM access token (optional)
        record_identifier (str): UID or title of the KSM record
        output_path (str): Path where to save the downloaded config
        
    Returns:
        bool: True if successful, False otherwise
    """
    if not check_ksm_dependency():
        return False
    
    # Validate file paths
    if ksm_config_path:
        is_valid, validated_config_path = validate_file_path(ksm_config_path)
        if not is_valid:
            print("ERROR: Invalid KSM config file path")
            return False
        ksm_config_path = validated_config_path
    
    is_valid, validated_output_path = validate_file_path(output_path)
    if not is_valid:
        print("ERROR: Invalid output file path")
        return False
    output_path = validated_output_path
    
    # Initialize SecretsManager
    secrets_manager = _initialize_secrets_manager(ksm_config_path, ksm_token)
    if not secrets_manager:
        return False
    
    try:
        # Get the record by UID or title
        secret = _get_secret_by_uid_or_title(secrets_manager, record_identifier)
        if not secret:
            return False
        
        # Find config.json attachment
        for file in secret.files:
            if file.name.lower() == 'config.json':
                # Ensure output directory exists
                os.makedirs(os.path.dirname(output_path), exist_ok=True)
                file.save_file(output_path, True)
                return True
        
        print("ERROR: config.json attachment not found in record")
        return False
        
    except Exception as e:
        print(f"ERROR: Failed to download config from KSM: {e}")
        return False


# =============================================================================
# FILE MANAGEMENT HELPER FUNCTIONS
# =============================================================================

def _get_file_uid(file_obj):
    """
    Extract file UID from file object.
    
    Args:
        file_obj: File object from KSM
        
    Returns:
        str: File UID or None if not found
    """
    try:
        # Try different ways to get the file UID
        if hasattr(file_obj, 'f') and file_obj.f:
            file_uid = file_obj.f.get('fileUid')
            if file_uid:
                return file_uid
        
        if hasattr(file_obj, 'fileUid'):
            return file_obj.fileUid
        
        if hasattr(file_obj, 'uid'):
            return file_obj.uid
            
        return None
        
    except Exception:
        return None

def _remove_existing_config_files(secrets_manager, secret, record_identifier):
    """
    Remove existing config.json files from KSM record.
    
    Args:
        secrets_manager: KSM SecretsManager instance
        secret: KSM secret object
        record_identifier: Record UID or title for refresh after removal
        
    Returns:
        tuple: (success, updated_secret)
    """
    try:
        # Find existing config.json files
        config_files = [
            f for f in secret.files if f.name.lower() == 'config.json'
        ]
        if not config_files:
            return True, secret
            
        files_to_remove = []
        for file_obj in config_files:
            file_uid = _get_file_uid(file_obj)
            if file_uid:
                files_to_remove.append(file_uid)
                print("Found config.json to remove UID: [REDACTED]")
            else:
                print(f"WARNING: Could not find UID for file: {file_obj.name}")
        
        if files_to_remove:
            secrets_manager.save(secret, links_to_remove=files_to_remove)
            print(f"Removed {len(files_to_remove)} config.json file(s)")
            # Refresh the secret after removal
            updated_secret = _get_secret_by_uid_or_title(
                secrets_manager, record_identifier
            )
            if updated_secret:
                return True, updated_secret
            else:
                print("WARNING: Could not refresh secret after file removal")
                return False, secret
        
        return True, secret
        
    except Exception as e:
        print(f"WARNING: Failed to remove existing files: {e}")
        return False, secret

def _upload_new_config_file(secrets_manager, secret, config_file_path):
    """
    Upload new config.json file to KSM record.
    
    Args:
        secrets_manager: KSM SecretsManager instance
        secret: KSM secret object
        config_file_path: Path to local config file
        
    Returns:
        bool: True if successful, False otherwise
    """
    from keeper_secrets_manager_core.core import KeeperFileUpload
    
    try:
        # Validate config file exists and is readable
        if not os.path.exists(config_file_path):
            print("ERROR: Config file not found")
            return False
        
        print("Uploading new config.json...")
        my_file = KeeperFileUpload.from_file(
            config_file_path, 'config.json', 'config.json'
        )
        secrets_manager.upload_file(secret, file=my_file)
        print("Successfully uploaded new config.json")
        return True
        
    except Exception as e:
        print(f"ERROR: Failed to upload config file: {e}")
        return False

def upload_config(ksm_config_path, ksm_token, record_identifier, config_file_path):
    """
    Upload config.json to KSM record, removing existing ones first.
    
    Args:
        ksm_config_path (str): Path to KSM config file (optional)
        ksm_token (str): KSM access token (optional)
        record_identifier (str): UID or title of the KSM record
        config_file_path (str): Path to local config file to upload
        
    Returns:
        bool: True if successful, False otherwise
    """
    if not check_ksm_dependency():
        return False
    
    # Validate file paths
    if ksm_config_path:
        is_valid, validated_config_path = validate_file_path(ksm_config_path)
        if not is_valid:
            print("ERROR: Invalid KSM config file path")
            return False
        ksm_config_path = validated_config_path
    
    is_valid, validated_config_file_path = validate_file_path(config_file_path)
    if not is_valid:
        print("ERROR: Invalid config file path")
        return False
    config_file_path = validated_config_file_path
    
    # Initialize SecretsManager
    secrets_manager = _initialize_secrets_manager(ksm_config_path, ksm_token)
    if not secrets_manager:
        return False
    
    try:
        # Get the record by UID or title
        secret = _get_secret_by_uid_or_title(secrets_manager, record_identifier)
        if not secret:
            return False
        
        # Remove existing config.json files
        success, updated_secret = _remove_existing_config_files(
            secrets_manager, secret, record_identifier
        )
        if not success:
            return False
        
        # Upload new config.json file
        return _upload_new_config_file(
            secrets_manager, updated_secret, config_file_path
        )
        
    except Exception as e:
        print(f"ERROR: Failed to upload config: {e}")
        return False


# =============================================================================
# FILE MONITORING AND UTILITIES
# =============================================================================

def _get_secure_file_hash(file_path):
    """
    Securely calculate file hash with proper error handling.
    
    Args:
        file_path (str): Path to file
        
    Returns:
        str: File hash or None if file doesn't exist/error
    """
    try:
        is_valid, validated_path = validate_file_path(file_path)
        if not is_valid:
            return None
        
        if not os.path.exists(validated_path):
            return None
            
        # Use context manager for atomic file read
        with open(validated_path, 'rb') as f:
            content = f.read()
            # Use SHA-256 instead of MD5 for security
            return hashlib.sha256(content).hexdigest()
            
    except (OSError, IOError):
        return None

def monitor_config(ksm_config_path, ksm_token, record_identifier, config_file_path):
    """
    Monitor config.json file for changes and upload when modified.
    
    Args:
        ksm_config_path (str): Path to KSM config file (optional)
        ksm_token (str): KSM access token (optional)
        record_identifier (str): UID or title of the KSM record
        config_file_path (str): Path to config file to monitor
    """
    # Validate file paths at startup
    is_valid, validated_config_file_path = validate_file_path(config_file_path)
    if not is_valid:
        print("ERROR: Invalid config file path for monitoring")
        return
    config_file_path = validated_config_file_path
    
    if ksm_config_path:
        is_valid, validated_ksm_config_path = validate_file_path(ksm_config_path)
        if not is_valid:
            print("ERROR: Invalid KSM config file path for monitoring")
            return
        ksm_config_path = validated_ksm_config_path
    
    print(f"Monitoring config file for changes")
    
    last_hash = _get_secure_file_hash(config_file_path)
    monitor_interval = 30  # Check every 30 seconds
    
    while True:
        try:
            time.sleep(monitor_interval)
            
            current_hash = _get_secure_file_hash(config_file_path)
            
            if current_hash is None:
                if last_hash is not None:
                    print("Config file was removed, continuing to monitor...")
                    last_hash = None
                continue
            
            if current_hash != last_hash:
                print("Config file changed, uploading to KSM record...")
                upload_success = upload_config(
                    ksm_config_path, ksm_token, record_identifier, config_file_path
                )
                if upload_success:
                    print("Config upload completed successfully")
                    # Only update hash on successful upload
                    last_hash = current_hash
                else:
                    print("Config upload failed, will retry on next change")
                
        except KeyboardInterrupt:
            print("Monitoring stopped by user")
            break
        except Exception as e:
            print(f"ERROR: Error in config monitor: {e}")
            time.sleep(5)


# =============================================================================
# COMMAND LINE INTERFACE
# =============================================================================

def _validate_command_args(args):
    """
    Validate command line arguments.
    
    Args:
        args: Parsed command line arguments
        
    Returns:
        bool: True if valid, False otherwise
    """
    # Handle process-config command separately
    if args.command == 'process-config':
        if not args.ksm_config:
            print("ERROR: --ksm-config is required for process-config command")
            return False
        return True
    
    # For other commands, validate required parameters
    if not args.record_uid:
        print("ERROR: --record-uid is required for this command")
        return False
    
    if not args.config_file:
        print("ERROR: --config-file is required for this command")
        return False
    
    # Validate authentication parameters
    if not args.ksm_config and not args.ksm_token:
        print("Either --ksm-config or --ksm-token must be provided")
        return False
    
    if args.ksm_config and args.ksm_token:
        print("Cannot specify both --ksm-config and --ksm-token")
        return False
    
    # Validate file paths early
    if args.ksm_config:
        is_valid, _ = validate_file_path(args.ksm_config)
        if not is_valid:
            print("Invalid KSM config file path")
            return False
    
    is_valid, _ = validate_file_path(args.config_file)
    if not is_valid:
        print("Invalid config file path")
        return False
    
    return True


def _execute_command(args):
    """
    Execute the specified command.
    
    Args:
        args: Parsed command line arguments
        
    Returns:
        bool: True if successful, False otherwise
    """
    if args.command == 'process-config':
        result_path = process_ksm_config(args.ksm_config)
        if result_path:
            print(result_path)  # Output the final config path
            return True
        return False
    
    elif args.command == 'download':
        return download_config(
            args.ksm_config, args.ksm_token, args.record_uid, args.config_file
        )
    
    elif args.command == 'upload':
        return upload_config(
            args.ksm_config, args.ksm_token, args.record_uid, args.config_file
        )
    
    elif args.command == 'monitor':
        # Monitor runs indefinitely
        monitor_config(
            args.ksm_config, args.ksm_token, args.record_uid, args.config_file
        )
        return True
    
    return False


def main():
    """
    Main entry point for the KSM Docker Utility.
    """
    parser = argparse.ArgumentParser(
        description="KSM Docker Utility - Secure file operations for KSM records",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    
    parser.add_argument(
        "command",
        choices=['download', 'upload', 'monitor', 'process-config'],
        help="Command to execute"
    )
    parser.add_argument(
        "--ksm-config",
        help="KSM config file path or base64-encoded JSON"
    )
    parser.add_argument(
        "--ksm-token",
        help="KSM access token"
    )
    parser.add_argument(
        "--record-uid",
        help="KSM record UID or title"
    )
    parser.add_argument(
        "--config-file",
        help="Local config.json file path"
    )
    
    args = parser.parse_args()
    
    # Validate arguments
    if not _validate_command_args(args):
        sys.exit(1)
    
    # Execute command
    success = False
    try:
        success = _execute_command(args)
    except KeyboardInterrupt:
        print("\nOperation interrupted by user")
        success = True
    except Exception as e:
        print(f"ERROR: Unexpected error occurred: {e}")
        success = False
    
    sys.exit(0 if success else 1)

if __name__ == "__main__":
    main()



import logging
from ... import utils, crypto
from ...params import KeeperParams


def is_addon_enabled(params, addon_name):    # type: (KeeperParams, Dict[str, ]) -> Boolean
    def is_enabled(addon):
        return addon.get('enabled') or addon.get('included_in_product')

    enterprise = params.enterprise or {}
    licenses = enterprise.get('licenses')
    if not isinstance(licenses, list):
        return False
    if next(iter(licenses), {}).get('lic_status') == 'business_trial':
        return True
    addons = [a for l in licenses for a in l.get('add_ons', []) if a.get('name') == addon_name]
    return any(a for a in addons if is_enabled(a))


def user_has_privilege(params, privilege):  # type: (KeeperParams, str) -> bool
    # Running as MSP admin, user has all available privileges in this context
    if params.msp_tree_key:
        return True

    enterprise = params.enterprise

    # Not an admin account (user has no admin privileges)
    if not enterprise:
        return False

    # Check role-derived privileges
    username = params.user
    users = enterprise.get('users')
    e_user_id = next(iter([u.get('enterprise_user_id') for u in users if u.get('username') == username]))
    role_users = enterprise.get('role_users')
    r_ids = [ru.get('role_id') for ru in role_users if ru.get('enterprise_user_id') == e_user_id]
    r_privileges = enterprise.get('role_privileges')
    p_key = 'privilege'
    return any(rp for rp in r_privileges if rp.get('role_id') in r_ids and rp.get(p_key) == privilege)

def get_enterprise_key(params, is_rsa=False):
    keys = params.enterprise.get('keys', {})
    try:
        pk_data = utils.base64_url_decode(keys.get(f'{is_rsa and "rsa" or "ecc"}_encrypted_private_key'))
        pk_data = crypto.decrypt_aes_v2(pk_data, params.enterprise['unencrypted_tree_key'])
        return is_rsa and crypto.load_rsa_private_key(pk_data) or crypto.load_ec_private_key(pk_data)
    except Exception as e:
        logging.debug(e)

def try_enterprise_decrypt(params, data):
    dec_params = [
        [crypto.decrypt_ec, get_enterprise_key(params, is_rsa=False)],
        [crypto.decrypt_rsa, get_enterprise_key(params, is_rsa=True)]
    ]
    while dec_params:
        decrypt_fn, pk = dec_params.pop()
        try:
            decrypted = decrypt_fn(data, pk)
            return decrypted
        except:
            continue
